@@ -24,9 +24,10 @@ package signaling
 import (
 	"encoding/json"
 	"fmt"
-	"reflect"
 	"sort"
 	"testing"
+
+	"github.com/stretchr/testify/assert"
 )
 
 type testCheckValid interface {
@@ -53,66 +54,62 @@ func wrapMessage(messageType string, msg testCheckValid) *ClientMessage {
 }
 
 func testMessages(t *testing.T, messageType string, valid_messages []testCheckValid, invalid_messages []testCheckValid) {
+	t.Helper()
+	assert := assert.New(t)
 	for _, msg := range valid_messages {
-		if err := msg.CheckValid(); err != nil {
-			t.Errorf("Message %+v should be valid, got %s", msg, err)
-		}
+		assert.NoError(msg.CheckValid(), "Message %+v should be valid", msg)
+
 		// If the inner message is valid, it should also be valid in a wrapped
 		// ClientMessage.
-		if wrapped := wrapMessage(messageType, msg); wrapped == nil {
-			t.Errorf("Unknown message type: %s", messageType)
-		} else if err := wrapped.CheckValid(); err != nil {
-			t.Errorf("Message %+v should be valid, got %s", wrapped, err)
+		if wrapped := wrapMessage(messageType, msg); assert.NotNil(wrapped, "Unknown message type: %s", messageType) {
+			assert.NoError(wrapped.CheckValid(), "Message %+v should be valid", wrapped)
 		}
 	}
 	for _, msg := range invalid_messages {
-		if err := msg.CheckValid(); err == nil {
-			t.Errorf("Message %+v should not be valid", msg)
-		}
+		assert.Error(msg.CheckValid(), "Message %+v should not be valid", msg)
 
 		// If the inner message is invalid, it should also be invalid in a
 		// wrapped ClientMessage.
-		if wrapped := wrapMessage(messageType, msg); wrapped == nil {
-			t.Errorf("Unknown message type: %s", messageType)
-		} else if err := wrapped.CheckValid(); err == nil {
-			t.Errorf("Message %+v should not be valid", wrapped)
+		if wrapped := wrapMessage(messageType, msg); assert.NotNil(wrapped, "Unknown message type: %s", messageType) {
+			assert.Error(wrapped.CheckValid(), "Message %+v should not be valid", wrapped)
 		}
 	}
 }
 
 func TestClientMessage(t *testing.T) {
+	t.Parallel()
+	assert := assert.New(t)
 	// The message needs a type.
 	msg := ClientMessage{}
-	if err := msg.CheckValid(); err == nil {
-		t.Errorf("Message %+v should not be valid", msg)
-	}
+	assert.Error(msg.CheckValid())
 }
 
 func TestHelloClientMessage(t *testing.T) {
+	t.Parallel()
 	internalAuthParams := []byte("{\"backend\":\"https://domain.invalid\"}")
 	tokenAuthParams := []byte("{\"token\":\"invalid-token\"}")
 	valid_messages := []testCheckValid{
 		// Hello version 1
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
-				Params: &json.RawMessage{'{', '}'},
+			Auth: &HelloClientMessageAuth{
+				Params: json.RawMessage("{}"),
 				Url:    "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   "client",
-				Params: &json.RawMessage{'{', '}'},
+				Params: json.RawMessage("{}"),
 				Url:    "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   "internal",
-				Params: (*json.RawMessage)(&internalAuthParams),
+				Params: internalAuthParams,
 			},
 		},
 		&HelloClientMessage{
@@ -122,16 +119,16 @@ func TestHelloClientMessage(t *testing.T) {
 		// Hello version 2
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
-				Params: (*json.RawMessage)(&tokenAuthParams),
+			Auth: &HelloClientMessageAuth{
+				Params: tokenAuthParams,
 				Url:    "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   "client",
-				Params: (*json.RawMessage)(&tokenAuthParams),
+				Params: tokenAuthParams,
 				Url:    "https://domain.invalid",
 			},
 		},
@@ -147,75 +144,75 @@ func TestHelloClientMessage(t *testing.T) {
 		&HelloClientMessage{Version: HelloVersionV1},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
-				Params: &json.RawMessage{'{', '}'},
+			Auth: &HelloClientMessageAuth{
+				Params: json.RawMessage("{}"),
 				Type:   "invalid-type",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Url: "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
-				Params: &json.RawMessage{'{', '}'},
+			Auth: &HelloClientMessageAuth{
+				Params: json.RawMessage("{}"),
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
-				Params: &json.RawMessage{'{', '}'},
+			Auth: &HelloClientMessageAuth{
+				Params: json.RawMessage("{}"),
 				Url:    "invalid-url",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   "internal",
-				Params: &json.RawMessage{'{', '}'},
+				Params: json.RawMessage("{}"),
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV1,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   "internal",
-				Params: &json.RawMessage{'x', 'y', 'z'}, // Invalid JSON.
+				Params: json.RawMessage("xyz"), // Invalid JSON.
 			},
 		},
 		// Hello version 2
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Url: "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
-				Params: (*json.RawMessage)(&tokenAuthParams),
+			Auth: &HelloClientMessageAuth{
+				Params: tokenAuthParams,
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
-				Params: (*json.RawMessage)(&tokenAuthParams),
+			Auth: &HelloClientMessageAuth{
+				Params: tokenAuthParams,
 				Url:    "invalid-url",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
-				Params: (*json.RawMessage)(&internalAuthParams),
+			Auth: &HelloClientMessageAuth{
+				Params: internalAuthParams,
 				Url:    "https://domain.invalid",
 			},
 		},
 		&HelloClientMessage{
 			Version: HelloVersionV2,
-			Auth: HelloClientMessageAuth{
-				Params: &json.RawMessage{'x', 'y', 'z'}, // Invalid JSON.
+			Auth: &HelloClientMessageAuth{
+				Params: json.RawMessage("xyz"), // Invalid JSON.
 				Url:    "https://domain.invalid",
 			},
 		},
@@ -227,32 +224,32 @@ func TestHelloClientMessage(t *testing.T) {
 	msg := ClientMessage{
 		Type: "hello",
 	}
-	if err := msg.CheckValid(); err == nil {
-		t.Errorf("Message %+v should not be valid", msg)
-	}
+	assert := assert.New(t)
+	assert.Error(msg.CheckValid())
 }
 
 func TestMessageClientMessage(t *testing.T) {
+	t.Parallel()
 	valid_messages := []testCheckValid{
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type:      "session",
 				SessionId: "the-session-id",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type:   "user",
 				UserId: "the-user-id",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type: "room",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 	}
 	invalid_messages := []testCheckValid{
@@ -267,20 +264,20 @@ func TestMessageClientMessage(t *testing.T) {
 			Recipient: MessageClientMessageRecipient{
 				Type: "session",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type:   "session",
 				UserId: "the-user-id",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type: "user",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
@@ -293,13 +290,13 @@ func TestMessageClientMessage(t *testing.T) {
 				Type:      "user",
 				SessionId: "the-user-id",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 		&MessageClientMessage{
 			Recipient: MessageClientMessageRecipient{
 				Type: "unknown-type",
 			},
-			Data: &json.RawMessage{'{', '}'},
+			Data: json.RawMessage("{}"),
 		},
 	}
 	testMessages(t, "message", valid_messages, invalid_messages)
@@ -308,12 +305,12 @@ func TestMessageClientMessage(t *testing.T) {
 	msg := ClientMessage{
 		Type: "message",
 	}
-	if err := msg.CheckValid(); err == nil {
-		t.Errorf("Message %+v should not be valid", msg)
-	}
+	assert := assert.New(t)
+	assert.Error(msg.CheckValid())
 }
 
 func TestByeClientMessage(t *testing.T) {
+	t.Parallel()
 	// Any "bye" message is valid.
 	valid_messages := []testCheckValid{
 		&ByeClientMessage{},
@@ -326,12 +323,12 @@ func TestByeClientMessage(t *testing.T) {
 	msg := ClientMessage{
 		Type: "bye",
 	}
-	if err := msg.CheckValid(); err != nil {
-		t.Errorf("Message %+v should be valid, got %s", msg, err)
-	}
+	assert := assert.New(t)
+	assert.NoError(msg.CheckValid())
 }
 
 func TestRoomClientMessage(t *testing.T) {
+	t.Parallel()
 	// Any "room" message is valid.
 	valid_messages := []testCheckValid{
 		&RoomClientMessage{},
@@ -344,66 +341,53 @@ func TestRoomClientMessage(t *testing.T) {
 	msg := ClientMessage{
 		Type: "room",
 	}
-	if err := msg.CheckValid(); err == nil {
-		t.Errorf("Message %+v should not be valid", msg)
-	}
+	assert := assert.New(t)
+	assert.Error(msg.CheckValid())
 }
 
 func TestErrorMessages(t *testing.T) {
+	t.Parallel()
+	assert := assert.New(t)
 	id := "request-id"
 	msg := ClientMessage{
 		Id: id,
 	}
 	err1 := msg.NewErrorServerMessage(&Error{})
-	if err1.Id != id {
-		t.Errorf("Expected id %s, got %+v", id, err1)
-	}
-	if err1.Type != "error" || err1.Error == nil {
-		t.Errorf("Expected type \"error\", got %+v", err1)
-	}
+	assert.Equal(id, err1.Id, "%+v", err1)
+	assert.Equal("error", err1.Type, "%+v", err1)
+	assert.NotNil(err1.Error, "%+v", err1)
 
 	err2 := msg.NewWrappedErrorServerMessage(fmt.Errorf("test-error"))
-	if err2.Id != id {
-		t.Errorf("Expected id %s, got %+v", id, err2)
-	}
-	if err2.Type != "error" || err2.Error == nil {
-		t.Errorf("Expected type \"error\", got %+v", err2)
-	}
-	if err2.Error.Code != "internal_error" {
-		t.Errorf("Expected code \"internal_error\", got %+v", err2)
-	}
-	if err2.Error.Message != "test-error" {
-		t.Errorf("Expected message \"test-error\", got %+v", err2)
+	assert.Equal(id, err2.Id, "%+v", err2)
+	assert.Equal("error", err2.Type, "%+v", err2)
+	if assert.NotNil(err2.Error, "%+v", err2) {
+		assert.Equal("internal_error", err2.Error.Code, "%+v", err2)
+		assert.Equal("test-error", err2.Error.Message, "%+v", err2)
 	}
 	// Test "error" interface
-	if err2.Error.Error() != "test-error" {
-		t.Errorf("Expected error string \"test-error\", got %+v", err2)
-	}
+	assert.Equal("test-error", err2.Error.Error(), "%+v", err2)
 }
 
 func TestIsChatRefresh(t *testing.T) {
+	t.Parallel()
 	var msg ServerMessage
 	data_true := []byte("{\"type\":\"chat\",\"chat\":{\"refresh\":true}}")
 	msg = ServerMessage{
 		Type: "message",
 		Message: &MessageServerMessage{
-			Data: (*json.RawMessage)(&data_true),
+			Data: data_true,
 		},
 	}
-	if !msg.IsChatRefresh() {
-		t.Error("message should be detected as chat refresh")
-	}
+	assert.True(t, msg.IsChatRefresh())
 
 	data_false := []byte("{\"type\":\"chat\",\"chat\":{\"refresh\":false}}")
 	msg = ServerMessage{
 		Type: "message",
 		Message: &MessageServerMessage{
-			Data: (*json.RawMessage)(&data_false),
+			Data: data_false,
 		},
 	}
-	if msg.IsChatRefresh() {
-		t.Error("message should not be detected as chat refresh")
-	}
+	assert.False(t, msg.IsChatRefresh())
 }
 
 func assertEqualStrings(t *testing.T, expected, result []string) {
@@ -420,26 +404,22 @@ func assertEqualStrings(t *testing.T, expected, result []string) {
 		sort.Strings(result)
 	}
 
-	if !reflect.DeepEqual(expected, result) {
-		t.Errorf("Expected %+v, got %+v", expected, result)
-	}
+	assert.Equal(t, expected, result)
 }
 
 func Test_Welcome_AddRemoveFeature(t *testing.T) {
+	t.Parallel()
+	assert := assert.New(t)
 	var msg WelcomeServerMessage
 	assertEqualStrings(t, []string{}, msg.Features)
 
 	msg.AddFeature("one", "two", "one")
 	assertEqualStrings(t, []string{"one", "two"}, msg.Features)
-	if !sort.StringsAreSorted(msg.Features) {
-		t.Errorf("features should be sorted, got %+v", msg.Features)
-	}
+	assert.True(sort.StringsAreSorted(msg.Features), "features should be sorted, got %+v", msg.Features)
 
 	msg.AddFeature("three")
 	assertEqualStrings(t, []string{"one", "two", "three"}, msg.Features)
-	if !sort.StringsAreSorted(msg.Features) {
-		t.Errorf("features should be sorted, got %+v", msg.Features)
-	}
+	assert.True(sort.StringsAreSorted(msg.Features), "features should be sorted, got %+v", msg.Features)
 
 	msg.RemoveFeature("three", "one")
 	assertEqualStrings(t, []string{"two"}, msg.Features)
