@@ -36,15 +36,16 @@ import (
 )
 
 var (
-	// Sessions expire 30 seconds after the connection closed.
-	sessionExpireDuration = 30 * time.Second
-
 	// Warn if a session has 32 or more pending messages.
 	warnPendingMessagesCount = 32
 
 	PathToOcsSignalingBackend = "ocs/v2.php/apps/spreed/api/v1/signaling/backend"
 )
 
+const (
+	FederatedRoomSessionIdPrefix = "federated|"
+)
+
 // ResponseHandlerFunc will return "true" has been fully processed.
 type ResponseHandlerFunc func(message *ClientMessage) bool
 
@@ -54,11 +55,15 @@ type ClientSession struct {
 	privateId string
 	publicId  string
 	data      *SessionIdData
+	ctx       context.Context
+	closeFunc context.CancelFunc
 
 	clientType string
 	features   []string
 	userId     string
-	userData   *json.RawMessage
+	userData   json.RawMessage
+
+	parseUserData func() (map[string]interface{}, error)
 
 	inCall              Flags
 	supportsPermissions bool
@@ -68,14 +73,15 @@ type ClientSession struct {
 	backendUrl       string
 	parsedBackendUrl *url.URL
 
-	expires time.Time
-
 	mu sync.Mutex
 
-	client        *Client
-	room          atomic.Pointer[Room]
-	roomJoinTime  atomic.Int64
-	roomSessionId string
+	client       HandlerClient
+	room         atomic.Pointer[Room]
+	roomJoinTime atomic.Int64
+	federation   atomic.Pointer[FederationClient]
+
+	roomSessionIdLock sync.RWMutex
+	roomSessionId     string
 
 	publisherWaiters ChannelWaiters
 
@@ -96,17 +102,21 @@ type ClientSession struct {
 }
 
 func NewClientSession(hub *Hub, privateId string, publicId string, data *SessionIdData, backend *Backend, hello *HelloClientMessage, auth *BackendClientAuthResponse) (*ClientSession, error) {
+	ctx, closeFunc := context.WithCancel(context.Background())
 	s := &ClientSession{
 		hub:       hub,
 		events:    hub.events,
 		privateId: privateId,
 		publicId:  publicId,
 		data:      data,
+		ctx:       ctx,
+		closeFunc: closeFunc,
 
-		clientType: hello.Auth.Type,
-		features:   hello.Features,
-		userId:     auth.UserId,
-		userData:   auth.User,
+		clientType:    hello.Auth.Type,
+		features:      hello.Features,
+		userId:        auth.UserId,
+		userData:      auth.User,
+		parseUserData: parseUserData(auth.User),
 
 		backend: backend,
 	}
@@ -145,6 +155,10 @@ func NewClientSession(hub *Hub, privateId string, publicId string, data *Session
 	return s, nil
 }
 
+func (s *ClientSession) Context() context.Context {
+	return s.ctx
+}
+
 func (s *ClientSession) PrivateId() string {
 	return s.privateId
 }
@@ -154,8 +168,8 @@ func (s *ClientSession) PublicId() string {
 }
 
 func (s *ClientSession) RoomSessionId() string {
-	s.mu.Lock()
-	defer s.mu.Unlock()
+	s.roomSessionIdLock.RLock()
+	defer s.roomSessionIdLock.RUnlock()
 	return s.roomSessionId
 }
 
@@ -309,28 +323,21 @@ func (s *ClientSession) UserId() string {
 	return userId
 }
 
-func (s *ClientSession) UserData() *json.RawMessage {
+func (s *ClientSession) UserData() json.RawMessage {
 	return s.userData
 }
 
-func (s *ClientSession) StartExpire() {
-	// The hub mutex must be held when calling this method.
-	s.expires = time.Now().Add(sessionExpireDuration)
-	s.hub.expiredSessions[s] = true
-}
-
-func (s *ClientSession) StopExpire() {
-	// The hub mutex must be held when calling this method.
-	delete(s.hub.expiredSessions, s)
-}
-
-func (s *ClientSession) IsExpired(now time.Time) bool {
-	return now.After(s.expires)
+func (s *ClientSession) ParsedUserData() (map[string]interface{}, error) {
+	return s.parseUserData()
 }
 
 func (s *ClientSession) SetRoom(room *Room) {
 	s.room.Store(room)
-	if room != nil {
+	s.onRoomSet(room != nil)
+}
+
+func (s *ClientSession) onRoomSet(hasRoom bool) {
+	if hasRoom {
 		s.roomJoinTime.Store(time.Now().UnixNano())
 	} else {
 		s.roomJoinTime.Store(0)
@@ -341,6 +348,22 @@ func (s *ClientSession) SetRoom(room *Room) {
 	s.seenJoinedEvents = nil
 }
 
+func (s *ClientSession) GetFederationClient() *FederationClient {
+	return s.federation.Load()
+}
+
+func (s *ClientSession) SetFederationClient(federation *FederationClient) {
+	s.mu.Lock()
+	defer s.mu.Unlock()
+
+	s.doLeaveRoom(true)
+	s.onRoomSet(federation != nil)
+
+	if prev := s.federation.Swap(federation); prev != nil && prev != federation {
+		prev.Close()
+	}
+}
+
 func (s *ClientSession) GetRoom() *Room {
 	return s.room.Load()
 }
@@ -357,7 +380,7 @@ func (s *ClientSession) getRoomJoinTime() time.Time {
 func (s *ClientSession) releaseMcuObjects() {
 	if len(s.publishers) > 0 {
 		go func(publishers map[StreamType]McuPublisher) {
-			ctx := context.TODO()
+			ctx := context.Background()
 			for _, publisher := range publishers {
 				publisher.Close(ctx)
 			}
@@ -366,7 +389,7 @@ func (s *ClientSession) releaseMcuObjects() {
 	}
 	if len(s.subscribers) > 0 {
 		go func(subscribers map[string]McuSubscriber) {
-			ctx := context.TODO()
+			ctx := context.Background()
 			for _, subscriber := range subscribers {
 				subscriber.Close(ctx)
 			}
@@ -380,8 +403,13 @@ func (s *ClientSession) Close() {
 }
 
 func (s *ClientSession) closeAndWait(wait bool) {
+	s.closeFunc()
 	s.hub.removeSession(s)
 
+	if prev := s.federation.Swap(nil); prev != nil {
+		prev.Close()
+	}
+
 	s.mu.Lock()
 	defer s.mu.Unlock()
 	if s.userId != "" {
@@ -413,8 +441,8 @@ func (s *ClientSession) SubscribeEvents() error {
 }
 
 func (s *ClientSession) UpdateRoomSessionId(roomSessionId string) error {
-	s.mu.Lock()
-	defer s.mu.Unlock()
+	s.roomSessionIdLock.Lock()
+	defer s.roomSessionIdLock.Unlock()
 
 	if s.roomSessionId == roomSessionId {
 		return nil
@@ -427,12 +455,16 @@ func (s *ClientSession) UpdateRoomSessionId(roomSessionId string) error {
 	if roomSessionId != "" {
 		if room := s.GetRoom(); room != nil {
 			log.Printf("Session %s updated room session id to %s in room %s", s.PublicId(), roomSessionId, room.Id())
+		} else if client := s.GetFederationClient(); client != nil {
+			log.Printf("Session %s updated room session id to %s in federated room %s", s.PublicId(), roomSessionId, client.RemoteRoomId())
 		} else {
 			log.Printf("Session %s updated room session id to %s in unknown room", s.PublicId(), roomSessionId)
 		}
 	} else {
 		if room := s.GetRoom(); room != nil {
 			log.Printf("Session %s cleared room session id in room %s", s.PublicId(), room.Id())
+		} else if client := s.GetFederationClient(); client != nil {
+			log.Printf("Session %s cleared room session id in federated room %s", s.PublicId(), client.RemoteRoomId())
 		} else {
 			log.Printf("Session %s cleared room session id in unknown room", s.PublicId())
 		}
@@ -443,8 +475,8 @@ func (s *ClientSession) UpdateRoomSessionId(roomSessionId string) error {
 }
 
 func (s *ClientSession) SubscribeRoomEvents(roomid string, roomSessionId string) error {
-	s.mu.Lock()
-	defer s.mu.Unlock()
+	s.roomSessionIdLock.Lock()
+	defer s.roomSessionIdLock.Unlock()
 
 	if err := s.events.RegisterRoomListener(roomid, s.backend, s); err != nil {
 		return err
@@ -475,9 +507,26 @@ func (s *ClientSession) LeaveCall() {
 }
 
 func (s *ClientSession) LeaveRoom(notify bool) *Room {
+	return s.LeaveRoomWithMessage(notify, nil)
+}
+
+func (s *ClientSession) LeaveRoomWithMessage(notify bool, message *ClientMessage) *Room {
+	if prev := s.federation.Swap(nil); prev != nil {
+		// Session was connected to a federation room.
+		if err := prev.Leave(message); err != nil {
+			log.Printf("Error leaving room for session %s on federation client %s: %s", s.PublicId(), prev.URL(), err)
+			prev.Close()
+		}
+		return nil
+	}
+
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
+	return s.doLeaveRoom(notify)
+}
+
+func (s *ClientSession) doLeaveRoom(notify bool) *Room {
 	room := s.GetRoom()
 	if room == nil {
 		return nil
@@ -503,11 +552,15 @@ func (s *ClientSession) doUnsubscribeRoomEvents(notify bool) {
 		s.events.UnregisterRoomListener(room.Id(), s.Backend(), s)
 	}
 	s.hub.roomSessions.DeleteRoomSession(s)
-	if notify && room != nil && s.roomSessionId != "" {
+
+	s.roomSessionIdLock.Lock()
+	defer s.roomSessionIdLock.Unlock()
+	if notify && room != nil && s.roomSessionId != "" && !strings.HasPrefix(s.roomSessionId, FederatedRoomSessionIdPrefix) {
 		// Notify
 		go func(sid string) {
 			ctx := context.Background()
 			request := NewBackendClientRoomRequest(room.Id(), s.userId, sid)
+			request.Room.UpdateFromSession(s)
 			request.Room.Action = "leave"
 			var response map[string]interface{}
 			if err := s.hub.backend.PerformJSONRequest(ctx, s.ParsedBackendUrl(), request, &response); err != nil {
@@ -520,14 +573,14 @@ func (s *ClientSession) doUnsubscribeRoomEvents(notify bool) {
 	s.roomSessionId = ""
 }
 
-func (s *ClientSession) ClearClient(client *Client) {
+func (s *ClientSession) ClearClient(client HandlerClient) {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
 	s.clearClientLocked(client)
 }
 
-func (s *ClientSession) clearClientLocked(client *Client) {
+func (s *ClientSession) clearClientLocked(client HandlerClient) {
 	if s.client == nil {
 		return
 	} else if client != nil && s.client != client {
@@ -540,18 +593,18 @@ func (s *ClientSession) clearClientLocked(client *Client) {
 	prevClient.SetSession(nil)
 }
 
-func (s *ClientSession) GetClient() *Client {
+func (s *ClientSession) GetClient() HandlerClient {
 	s.mu.Lock()
 	defer s.mu.Unlock()
 
 	return s.getClientUnlocked()
 }
 
-func (s *ClientSession) getClientUnlocked() *Client {
+func (s *ClientSession) getClientUnlocked() HandlerClient {
 	return s.client
 }
 
-func (s *ClientSession) SetClient(client *Client) *Client {
+func (s *ClientSession) SetClient(client HandlerClient) HandlerClient {
 	if client == nil {
 		panic("Use ClearClient to set the client to nil")
 	}
@@ -594,7 +647,7 @@ func (s *ClientSession) sendOffer(client McuClient, sender string, streamType St
 				Type:      "session",
 				SessionId: sender,
 			},
-			Data: (*json.RawMessage)(&offer_data),
+			Data: offer_data,
 		},
 	}
 
@@ -624,7 +677,7 @@ func (s *ClientSession) sendCandidate(client McuClient, sender string, streamTyp
 				Type:      "session",
 				SessionId: sender,
 			},
-			Data: (*json.RawMessage)(&candidate_data),
+			Data: candidate_data,
 		},
 	}
 
@@ -740,23 +793,6 @@ func (s *ClientSession) SubscriberClosed(subscriber McuSubscriber) {
 	}
 }
 
-type SdpError struct {
-	message string
-}
-
-func (e *SdpError) Error() string {
-	return e.message
-}
-
-type WrappedSdpError struct {
-	SdpError
-	err error
-}
-
-func (e *WrappedSdpError) Unwrap() error {
-	return e.err
-}
-
 type PermissionError struct {
 	permission Permission
 }
@@ -769,23 +805,10 @@ func (e *PermissionError) Error() string {
 	return fmt.Sprintf("permission \"%s\" not found", e.permission)
 }
 
-func (s *ClientSession) isSdpAllowedToSendLocked(payload map[string]interface{}) (MediaType, error) {
-	sdpValue, found := payload["sdp"]
-	if !found {
-		return 0, &SdpError{"payload does not contain a sdp"}
-	}
-	sdpText, ok := sdpValue.(string)
-	if !ok {
-		return 0, &SdpError{"payload does not contain a valid sdp"}
-	}
-	var sdp sdp.SessionDescription
-	if err := sdp.Unmarshal([]byte(sdpText)); err != nil {
-		return 0, &WrappedSdpError{
-			SdpError: SdpError{
-				message: fmt.Sprintf("could not parse sdp: %s", err),
-			},
-			err: err,
-		}
+func (s *ClientSession) isSdpAllowedToSendLocked(sdp *sdp.SessionDescription) (MediaType, error) {
+	if sdp == nil {
+		// Should have already been checked when data was validated.
+		return 0, ErrNoSdp
 	}
 
 	var mediaTypes MediaType
@@ -823,8 +846,8 @@ func (s *ClientSession) IsAllowedToSend(data *MessageClientMessageData) error {
 		// Client is allowed to publish any media (audio / video).
 		return nil
 	} else if data != nil && data.Type == "offer" {
-		// Parse SDP to check what user is trying to publish and check permissions accordingly.
-		if _, err := s.isSdpAllowedToSendLocked(data.Payload); err != nil {
+		// Check what user is trying to publish and check permissions accordingly.
+		if _, err := s.isSdpAllowedToSendLocked(data.offerSdp); err != nil {
 			return err
 		}
 
@@ -854,7 +877,7 @@ func (s *ClientSession) checkOfferTypeLocked(streamType StreamType, data *Messag
 
 		return MediaTypeScreen, nil
 	} else if data != nil && data.Type == "offer" {
-		mediaTypes, err := s.isSdpAllowedToSendLocked(data.Payload)
+		mediaTypes, err := s.isSdpAllowedToSendLocked(data.offerSdp)
 		if err != nil {
 			return 0, err
 		}
@@ -880,7 +903,15 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 		s.mu.Unlock()
 		defer s.mu.Lock()
 
-		bitrate := data.Bitrate
+		settings := NewPublisherSettings{
+			Bitrate:    data.Bitrate,
+			MediaTypes: mediaTypes,
+
+			AudioCodec:  data.AudioCodec,
+			VideoCodec:  data.VideoCodec,
+			VP9Profile:  data.VP9Profile,
+			H264Profile: data.H264Profile,
+		}
 		if backend := s.Backend(); backend != nil {
 			var maxBitrate int
 			if streamType == StreamTypeScreen {
@@ -888,14 +919,14 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 			} else {
 				maxBitrate = backend.maxStreamBitrate
 			}
-			if bitrate <= 0 {
-				bitrate = maxBitrate
-			} else if maxBitrate > 0 && bitrate > maxBitrate {
-				bitrate = maxBitrate
+			if settings.Bitrate <= 0 {
+				settings.Bitrate = maxBitrate
+			} else if maxBitrate > 0 && settings.Bitrate > maxBitrate {
+				settings.Bitrate = maxBitrate
 			}
 		}
 		var err error
-		publisher, err = mcu.NewPublisher(ctx, s, s.PublicId(), data.Sid, streamType, bitrate, mediaTypes, client)
+		publisher, err = mcu.NewPublisher(ctx, s, s.PublicId(), data.Sid, streamType, settings, client)
 		if err != nil {
 			return nil, err
 		}
@@ -905,7 +936,7 @@ func (s *ClientSession) GetOrCreatePublisher(ctx context.Context, mcu Mcu, strea
 		if prev, found := s.publishers[streamType]; found {
 			// Another thread created the publisher while we were waiting.
 			go func(pub McuPublisher) {
-				closeCtx := context.TODO()
+				closeCtx := context.Background()
 				pub.Close(closeCtx)
 			}(publisher)
 			publisher = prev
@@ -969,9 +1000,10 @@ func (s *ClientSession) GetOrCreateSubscriber(ctx context.Context, mcu Mcu, id s
 
 	subscriber, found := s.subscribers[getStreamId(id, streamType)]
 	if !found {
+		client := s.getClientUnlocked()
 		s.mu.Unlock()
 		var err error
-		subscriber, err = mcu.NewSubscriber(ctx, s, id, streamType)
+		subscriber, err = mcu.NewSubscriber(ctx, s, id, streamType, client)
 		s.mu.Lock()
 		if err != nil {
 			return nil, err
@@ -982,7 +1014,7 @@ func (s *ClientSession) GetOrCreateSubscriber(ctx context.Context, mcu Mcu, id s
 		if prev, found := s.subscribers[getStreamId(id, streamType)]; found {
 			// Another thread created the subscriber while we were waiting.
 			go func(sub McuSubscriber) {
-				closeCtx := context.TODO()
+				closeCtx := context.Background()
 				sub.Close(closeCtx)
 			}(subscriber)
 			subscriber = prev
@@ -1056,7 +1088,7 @@ func (s *ClientSession) processAsyncMessage(message *AsyncMessage) {
 	case "sendoffer":
 		// Process asynchronously to not block other messages received.
 		go func() {
-			ctx, cancel := context.WithTimeout(context.Background(), s.hub.mcuTimeout)
+			ctx, cancel := context.WithTimeout(s.Context(), s.hub.mcuTimeout)
 			defer cancel()
 
 			mc, err := s.GetOrCreateSubscriber(ctx, s.hub.mcu, message.SendOffer.SessionId, StreamType(message.SendOffer.Data.RoomType))
@@ -1088,7 +1120,7 @@ func (s *ClientSession) processAsyncMessage(message *AsyncMessage) {
 				return
 			}
 
-			mc.SendMessage(context.TODO(), nil, message.SendOffer.Data, func(err error, response map[string]interface{}) {
+			mc.SendMessage(s.Context(), nil, message.SendOffer.Data, func(err error, response map[string]interface{}) {
 				if err != nil {
 					log.Printf("Could not send MCU message %+v for session %s to %s: %s", message.SendOffer.Data, message.SendOffer.SessionId, s.PublicId(), err)
 					if err := s.events.PublishSessionMessage(message.SendOffer.SessionId, s.backend, &AsyncMessage{
@@ -1146,13 +1178,13 @@ func (s *ClientSession) storePendingMessage(message *ServerMessage) {
 func filterDisplayNames(events []*EventServerMessageSessionEntry) []*EventServerMessageSessionEntry {
 	result := make([]*EventServerMessageSessionEntry, 0, len(events))
 	for _, event := range events {
-		if event.User == nil {
+		if len(event.User) == 0 {
 			result = append(result, event)
 			continue
 		}
 
 		var userdata map[string]interface{}
-		if err := json.Unmarshal(*event.User, &userdata); err != nil {
+		if err := json.Unmarshal(event.User, &userdata); err != nil {
 			result = append(result, event)
 			continue
 		}
@@ -1178,7 +1210,7 @@ func filterDisplayNames(events []*EventServerMessageSessionEntry) []*EventServer
 		}
 
 		e := event.Clone()
-		e.User = (*json.RawMessage)(&data)
+		e.User = data
 		result = append(result, e)
 	}
 	return result
@@ -1273,12 +1305,12 @@ func (s *ClientSession) filterMessage(message *ServerMessage) *ServerMessage {
 					delete(s.seenJoinedEvents, e)
 				}
 			case "message":
-				if message.Event.Message == nil || message.Event.Message.Data == nil || len(*message.Event.Message.Data) == 0 || !s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
+				if message.Event.Message == nil || len(message.Event.Message.Data) == 0 || !s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
 					return message
 				}
 
 				var data RoomEventMessageData
-				if err := json.Unmarshal(*message.Event.Message.Data, &data); err != nil {
+				if err := json.Unmarshal(message.Event.Message.Data, &data); err != nil {
 					return message
 				}
 
@@ -1295,7 +1327,7 @@ func (s *ClientSession) filterMessage(message *ServerMessage) *ServerMessage {
 									Target: message.Event.Target,
 									Message: &RoomEventMessage{
 										RoomId: message.Event.Message.RoomId,
-										Data:   (*json.RawMessage)(&encoded),
+										Data:   encoded,
 									},
 								},
 							}
@@ -1305,9 +1337,9 @@ func (s *ClientSession) filterMessage(message *ServerMessage) *ServerMessage {
 			}
 		}
 	case "message":
-		if message.Message != nil && message.Message.Data != nil && len(*message.Message.Data) > 0 && s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
+		if message.Message != nil && len(message.Message.Data) > 0 && s.HasPermission(PERMISSION_HIDE_DISPLAYNAMES) {
 			var data MessageServerMessageData
-			if err := json.Unmarshal(*message.Message.Data, &data); err != nil {
+			if err := json.Unmarshal(message.Message.Data, &data); err != nil {
 				return message
 			}
 
@@ -1330,18 +1362,34 @@ func (s *ClientSession) filterAsyncMessage(msg *AsyncMessage) *ServerMessage {
 
 		switch msg.Message.Type {
 		case "message":
-			if msg.Message.Message != nil &&
-				msg.Message.Message.Sender != nil &&
-				msg.Message.Message.Sender.SessionId == s.PublicId() {
-				// Don't send message back to sender (can happen if sent to user or room)
-				return nil
+			if msg.Message.Message != nil {
+				if sender := msg.Message.Message.Sender; sender != nil {
+					if sender.SessionId == s.PublicId() {
+						// Don't send message back to sender (can happen if sent to user or room)
+						return nil
+					}
+					if sender.Type == RecipientTypeCall {
+						if room := s.GetRoom(); room == nil || !room.IsSessionInCall(s) {
+							// Session is not in call, so discard.
+							return nil
+						}
+					}
+				}
 			}
 		case "control":
-			if msg.Message.Control != nil &&
-				msg.Message.Control.Sender != nil &&
-				msg.Message.Control.Sender.SessionId == s.PublicId() {
-				// Don't send message back to sender (can happen if sent to user or room)
-				return nil
+			if msg.Message.Control != nil {
+				if sender := msg.Message.Control.Sender; sender != nil {
+					if sender.SessionId == s.PublicId() {
+						// Don't send message back to sender (can happen if sent to user or room)
+						return nil
+					}
+					if sender.Type == RecipientTypeCall {
+						if room := s.GetRoom(); room == nil || !room.IsSessionInCall(s) {
+							// Session is not in call, so discard.
+							return nil
+						}
+					}
+				}
 			}
 		case "event":
 			if msg.Message.Event.Target == "room" {
@@ -1361,7 +1409,7 @@ func (s *ClientSession) filterAsyncMessage(msg *AsyncMessage) *ServerMessage {
 	}
 }
 
-func (s *ClientSession) NotifySessionResumed(client *Client) {
+func (s *ClientSession) NotifySessionResumed(client HandlerClient) {
 	s.mu.Lock()
 	if len(s.pendingClientMessages) == 0 {
 		s.mu.Unlock()
