@@ -24,6 +24,8 @@ package signaling
 import (
 	"testing"
 	"time"
+
+	"github.com/stretchr/testify/assert"
 )
 
 func TestDeferredExecutor_MultiClose(t *testing.T) {
@@ -35,6 +37,7 @@ func TestDeferredExecutor_MultiClose(t *testing.T) {
 }
 
 func TestDeferredExecutor_QueueSize(t *testing.T) {
+	t.Parallel()
 	e := NewDeferredExecutor(0)
 	defer e.waitForStop()
 	defer e.Close()
@@ -52,9 +55,7 @@ func TestDeferredExecutor_QueueSize(t *testing.T) {
 	b := time.Now()
 	delta := b.Sub(a)
 	// Allow one millisecond less delay to account for time variance on CI runners.
-	if delta+time.Millisecond < delay {
-		t.Errorf("Expected a delay of %s, got %s", delay, delta)
-	}
+	assert.GreaterOrEqual(t, delta+time.Millisecond, delay)
 }
 
 func TestDeferredExecutor_Order(t *testing.T) {
@@ -80,9 +81,7 @@ func TestDeferredExecutor_Order(t *testing.T) {
 	<-done
 
 	for x := 0; x < 10; x++ {
-		if entries[x] != x {
-			t.Errorf("Expected %d at position %d, got %d", x, x, entries[x])
-		}
+		assert.Equal(t, entries[x], x, "Unexpected at position %d", x)
 	}
 }
 
@@ -100,13 +99,14 @@ func TestDeferredExecutor_CloseFromFunc(t *testing.T) {
 }
 
 func TestDeferredExecutor_DeferAfterClose(t *testing.T) {
+	CatchLogForTest(t)
 	e := NewDeferredExecutor(64)
 	defer e.waitForStop()
 
 	e.Close()
 
 	e.Execute(func() {
-		t.Error("method should not have been called")
+		assert.Fail(t, "method should not have been called")
 	})
 }
 
