@@ -44,6 +44,22 @@ if [ ! -f "$CONFIG" ]; then
     sed -i "s|#country =.*|country = $COUNTRY|" "$CONFIG"
   fi
 
+  if [ -n "$EXTERNAL_HOSTNAME" ]; then
+    sed -i "s|#hostname =.*|hostname = $EXTERNAL_HOSTNAME|" "$CONFIG"
+  fi
+  if [ -n "$TOKEN_ID" ]; then
+    sed -i "s|#token_id =.*|token_id = $TOKEN_ID|" "$CONFIG"
+  fi
+  if [ -n "$TOKEN_KEY" ]; then
+    sed -i "s|#token_key =.*|token_key = $TOKEN_KEY|" "$CONFIG"
+  fi
+  if [ -n "$BANDWIDTH_INCOMING" ]; then
+    sed -i "s|#incoming =.*|incoming = $BANDWIDTH_INCOMING|" "$CONFIG"
+  fi
+  if [ -n "$BANDWIDTH_OUTGOING" ]; then
+    sed -i "s|#outgoing =.*|outgoing = $BANDWIDTH_OUTGOING|" "$CONFIG"
+  fi
+
   HAS_ETCD=
   if [ -n "$ETCD_ENDPOINTS" ]; then
     sed -i "s|#endpoints =.*|endpoints = $ETCD_ENDPOINTS|" "$CONFIG"
@@ -109,7 +125,17 @@ if [ ! -f "$CONFIG" ]; then
   if [ -n "$STATS_IPS" ]; then
     sed -i "s|#allowed_ips =.*|allowed_ips = $STATS_IPS|" "$CONFIG"
   fi
+
+  if [ -n "$TRUSTED_PROXIES" ]; then
+    sed -i "s|#trustedproxies =.*|trustedproxies = $TRUSTED_PROXIES|" "$CONFIG"
+  fi
+fi
+
+if [ -n "$CA_CERTIFICATES_FOLDER" ]; then
+  echo "Adding files from $CA_CERTIFICATES_FOLDER to the CA store ..."
+  cp "$CA_CERTIFICATES_FOLDER"/* /usr/local/share/ca-certificates/
+  update-ca-certificates
 fi
 
 echo "Starting signaling proxy with $CONFIG ..."
-exec /usr/bin/nextcloud-spreed-signaling-proxy -config "$CONFIG"
+exec su-exec spreedbackend /usr/bin/nextcloud-spreed-signaling-proxy -config "$CONFIG"
