@@ -219,6 +219,17 @@ type dummyGatewayListener struct {
 func (l *dummyGatewayListener) ConnectionInterrupted() {
 }
 
+type JanusGatewayInterface interface {
+	Info(context.Context) (*InfoMsg, error)
+	Create(context.Context) (*JanusSession, error)
+	Close() error
+
+	send(map[string]interface{}, *transaction) (uint64, error)
+	removeTransaction(uint64)
+
+	removeSession(*JanusSession)
+}
+
 // Gateway represents a connection to an instance of the Janus Gateway.
 type JanusGateway struct {
 	listener GatewayListener
@@ -258,8 +269,8 @@ type JanusGateway struct {
 // 	return gateway, nil
 // }
 
-func NewJanusGateway(wsURL string, listener GatewayListener) (*JanusGateway, error) {
-	conn, _, err := janusDialer.Dial(wsURL, nil)
+func NewJanusGateway(ctx context.Context, wsURL string, listener GatewayListener) (*JanusGateway, error) {
+	conn, _, err := janusDialer.DialContext(ctx, wsURL, nil)
 	if err != nil {
 		return nil, err
 	}
@@ -310,7 +321,7 @@ func (gateway *JanusGateway) cancelTransactions() {
 			t.quit()
 		}(t)
 	}
-	gateway.transactions = make(map[uint64]*transaction)
+	clear(gateway.transactions)
 	gateway.Unlock()
 }
 
@@ -560,12 +571,18 @@ func (gateway *JanusGateway) Create(ctx context.Context) (*JanusSession, error)
 
 	// Store this session
 	gateway.Lock()
+	defer gateway.Unlock()
 	gateway.Sessions[session.Id] = session
-	gateway.Unlock()
 
 	return session, nil
 }
 
+func (gateway *JanusGateway) removeSession(session *JanusSession) {
+	gateway.Lock()
+	defer gateway.Unlock()
+	delete(gateway.Sessions, session.Id)
+}
+
 // Session represents a session instance on the Janus Gateway.
 type JanusSession struct {
 	// Id is the session_id of this session
@@ -578,7 +595,7 @@ type JanusSession struct {
 	// and Session.Unlock() methods provided by the embedded sync.Mutex.
 	sync.Mutex
 
-	gateway *JanusGateway
+	gateway JanusGatewayInterface
 }
 
 func (session *JanusSession) send(msg map[string]interface{}, t *transaction) (uint64, error) {
@@ -670,9 +687,7 @@ func (session *JanusSession) Destroy(ctx context.Context) (*janus.AckMsg, error)
 	}
 
 	// Remove this session from the gateway
-	session.gateway.Lock()
-	delete(session.gateway.Sessions, session.Id)
-	session.gateway.Unlock()
+	session.gateway.removeSession(session)
 
 	return ack, nil
 }
