@@ -33,6 +33,7 @@ import (
 	"net/http"
 	"net/url"
 	"os"
+	"slices"
 	"sort"
 	"strconv"
 	"strings"
@@ -41,7 +42,7 @@ import (
 	"time"
 
 	"github.com/dlintw/goconf"
-	"github.com/golang-jwt/jwt/v4"
+	"github.com/golang-jwt/jwt/v5"
 	"github.com/gorilla/websocket"
 )
 
@@ -133,11 +134,11 @@ func (c *mcuProxyPubSubCommon) doProcessPayload(client McuClient, msg *PayloadPr
 type mcuProxyPublisher struct {
 	mcuProxyPubSubCommon
 
-	id         string
-	mediaTypes MediaType
+	id       string
+	settings NewPublisherSettings
 }
 
-func newMcuProxyPublisher(id string, sid string, streamType StreamType, maxBitrate int, mediaTypes MediaType, proxyId string, conn *mcuProxyConnection, listener McuListener) *mcuProxyPublisher {
+func newMcuProxyPublisher(id string, sid string, streamType StreamType, maxBitrate int, settings NewPublisherSettings, proxyId string, conn *mcuProxyConnection, listener McuListener) *mcuProxyPublisher {
 	return &mcuProxyPublisher{
 		mcuProxyPubSubCommon: mcuProxyPubSubCommon{
 			sid:        sid,
@@ -147,21 +148,22 @@ func newMcuProxyPublisher(id string, sid string, streamType StreamType, maxBitra
 			conn:       conn,
 			listener:   listener,
 		},
-		id:         id,
-		mediaTypes: mediaTypes,
+		id:       id,
+		settings: settings,
 	}
 }
 
 func (p *mcuProxyPublisher) HasMedia(mt MediaType) bool {
-	return (p.mediaTypes & mt) == mt
+	return (p.settings.MediaTypes & mt) == mt
 }
 
 func (p *mcuProxyPublisher) SetMedia(mt MediaType) {
 	// TODO: Also update mediaTypes on proxy.
-	p.mediaTypes = mt
+	p.settings.MediaTypes = mt
 }
 
 func (p *mcuProxyPublisher) NotifyClosed() {
+	log.Printf("Publisher %s at %s was closed", p.proxyId, p.conn)
 	p.listener.PublisherClosed(p)
 	p.conn.removePublisher(p)
 }
@@ -185,7 +187,7 @@ func (p *mcuProxyPublisher) Close(ctx context.Context) {
 		return
 	}
 
-	log.Printf("Delete publisher %s at %s", p.proxyId, p.conn)
+	log.Printf("Deleted publisher %s at %s", p.proxyId, p.conn)
 }
 
 func (p *mcuProxyPublisher) SendMessage(ctx context.Context, message *MessageClientMessage, data *MessageClientMessageData, callback func(error, map[string]interface{})) {
@@ -217,13 +219,26 @@ func (p *mcuProxyPublisher) ProcessEvent(msg *EventProxyServerMessage) {
 	}
 }
 
+func (p *mcuProxyPublisher) GetStreams(ctx context.Context) ([]PublisherStream, error) {
+	return nil, errors.New("not implemented")
+}
+
+func (p *mcuProxyPublisher) PublishRemote(ctx context.Context, remoteId string, hostname string, port int, rtcpPort int) error {
+	return errors.New("remote publishing not supported for proxy publishers")
+}
+
+func (p *mcuProxyPublisher) UnpublishRemote(ctx context.Context, remoteId string, hostname string, port int, rtcpPort int) error {
+	return errors.New("remote publishing not supported for proxy publishers")
+}
+
 type mcuProxySubscriber struct {
 	mcuProxyPubSubCommon
 
-	publisherId string
+	publisherId   string
+	publisherConn *mcuProxyConnection
 }
 
-func newMcuProxySubscriber(publisherId string, sid string, streamType StreamType, maxBitrate int, proxyId string, conn *mcuProxyConnection, listener McuListener) *mcuProxySubscriber {
+func newMcuProxySubscriber(publisherId string, sid string, streamType StreamType, maxBitrate int, proxyId string, conn *mcuProxyConnection, listener McuListener, publisherConn *mcuProxyConnection) *mcuProxySubscriber {
 	return &mcuProxySubscriber{
 		mcuProxyPubSubCommon: mcuProxyPubSubCommon{
 			sid:        sid,
@@ -234,7 +249,8 @@ func newMcuProxySubscriber(publisherId string, sid string, streamType StreamType
 			listener:   listener,
 		},
 
-		publisherId: publisherId,
+		publisherId:   publisherId,
+		publisherConn: publisherConn,
 	}
 }
 
@@ -243,6 +259,11 @@ func (s *mcuProxySubscriber) Publisher() string {
 }
 
 func (s *mcuProxySubscriber) NotifyClosed() {
+	if s.publisherConn != nil {
+		log.Printf("Remote subscriber %s at %s (forwarded to %s) was closed", s.proxyId, s.conn, s.publisherConn)
+	} else {
+		log.Printf("Subscriber %s at %s was closed", s.proxyId, s.conn)
+	}
 	s.listener.SubscriberClosed(s)
 	s.conn.removeSubscriber(s)
 }
@@ -259,14 +280,26 @@ func (s *mcuProxySubscriber) Close(ctx context.Context) {
 	}
 
 	if response, err := s.conn.performSyncRequest(ctx, msg); err != nil {
-		log.Printf("Could not delete subscriber %s at %s: %s", s.proxyId, s.conn, err)
+		if s.publisherConn != nil {
+			log.Printf("Could not delete remote subscriber %s at %s (forwarded to %s): %s", s.proxyId, s.conn, s.publisherConn, err)
+		} else {
+			log.Printf("Could not delete subscriber %s at %s: %s", s.proxyId, s.conn, err)
+		}
 		return
 	} else if response.Type == "error" {
-		log.Printf("Could not delete subscriber %s at %s: %s", s.proxyId, s.conn, response.Error)
+		if s.publisherConn != nil {
+			log.Printf("Could not delete remote subscriber %s at %s (forwarded to %s): %s", s.proxyId, s.conn, s.publisherConn, response.Error)
+		} else {
+			log.Printf("Could not delete subscriber %s at %s: %s", s.proxyId, s.conn, response.Error)
+		}
 		return
 	}
 
-	log.Printf("Delete subscriber %s at %s", s.proxyId, s.conn)
+	if s.publisherConn != nil {
+		log.Printf("Deleted remote subscriber %s at %s (forwarded to %s)", s.proxyId, s.conn, s.publisherConn)
+	} else {
+		log.Printf("Deleted subscriber %s at %s", s.proxyId, s.conn)
+	}
 }
 
 func (s *mcuProxySubscriber) SendMessage(ctx context.Context, message *MessageClientMessage, data *MessageClientMessageData, callback func(error, map[string]interface{})) {
@@ -308,6 +341,7 @@ type mcuProxyConnection struct {
 	ip     net.IP
 
 	load       atomic.Int64
+	bandwidth  atomic.Pointer[EventProxyServerBandwidth]
 	mu         sync.Mutex
 	closer     *Closer
 	closedDone *Closer
@@ -326,7 +360,7 @@ type mcuProxyConnection struct {
 
 	msgId      atomic.Int64
 	helloMsgId string
-	sessionId  string
+	sessionId  atomic.Value
 	country    atomic.Value
 
 	callbacks map[string]func(*ProxyServerMessage)
@@ -359,6 +393,7 @@ func newMcuProxyConnection(proxy *mcuProxy, baseUrl string, ip net.IP) (*mcuProx
 	}
 	conn.reconnectInterval.Store(int64(initialReconnectInterval))
 	conn.load.Store(loadNotConnected)
+	conn.bandwidth.Store(nil)
 	conn.country.Store("")
 	return conn, nil
 }
@@ -371,6 +406,54 @@ func (c *mcuProxyConnection) String() string {
 	return c.rawUrl
 }
 
+func (c *mcuProxyConnection) IsSameCountry(initiator McuInitiator) bool {
+	if initiator == nil {
+		return true
+	}
+
+	initiatorCountry := initiator.Country()
+	if initiatorCountry == "" {
+		return true
+	}
+
+	connCountry := c.Country()
+	if connCountry == "" {
+		return true
+	}
+
+	return initiatorCountry == connCountry
+}
+
+func (c *mcuProxyConnection) IsSameContinent(initiator McuInitiator) bool {
+	if initiator == nil {
+		return true
+	}
+
+	initiatorCountry := initiator.Country()
+	if initiatorCountry == "" {
+		return true
+	}
+
+	connCountry := c.Country()
+	if connCountry == "" {
+		return true
+	}
+
+	initiatorContinents, found := ContinentMap[initiatorCountry]
+	if found {
+		m := c.proxy.getContinentsMap()
+		// Map continents to other continents (e.g. use Europe for Africa).
+		for _, continent := range initiatorContinents {
+			if toAdd, found := m[continent]; found {
+				initiatorContinents = append(initiatorContinents, toAdd...)
+			}
+		}
+
+	}
+	connContinents := ContinentMap[connCountry]
+	return ContinentsOverlap(initiatorContinents, connContinents)
+}
+
 type mcuProxyConnectionStats struct {
 	Url        string     `json:"url"`
 	IP         net.IP     `json:"ip,omitempty"`
@@ -414,10 +497,29 @@ func (c *mcuProxyConnection) Load() int64 {
 	return c.load.Load()
 }
 
+func (c *mcuProxyConnection) Bandwidth() *EventProxyServerBandwidth {
+	return c.bandwidth.Load()
+}
+
 func (c *mcuProxyConnection) Country() string {
 	return c.country.Load().(string)
 }
 
+func (c *mcuProxyConnection) SessionId() string {
+	sid := c.sessionId.Load()
+	if sid == nil {
+		return ""
+	}
+
+	return sid.(string)
+}
+
+func (c *mcuProxyConnection) IsConnected() bool {
+	c.mu.Lock()
+	defer c.mu.Unlock()
+	return c.conn != nil && c.SessionId() != ""
+}
+
 func (c *mcuProxyConnection) IsTemporary() bool {
 	return c.temporary.Load()
 }
@@ -443,7 +545,10 @@ func (c *mcuProxyConnection) readPump() {
 		}
 	}()
 	defer c.close()
-	defer c.load.Store(loadNotConnected)
+	defer func() {
+		c.load.Store(loadNotConnected)
+		c.bandwidth.Store(nil)
+	}()
 
 	c.mu.Lock()
 	conn := c.conn
@@ -744,8 +849,9 @@ func (c *mcuProxyConnection) clearPublishers() {
 			publisher.NotifyClosed()
 		}
 	}(c.publishers)
+	// Can't use clear(...) here as the map is processed by the goroutine above.
 	c.publishers = make(map[string]*mcuProxyPublisher)
-	c.publisherIds = make(map[string]string)
+	clear(c.publisherIds)
 
 	if c.closeScheduled.Load() || c.IsTemporary() {
 		go c.closeIfEmpty()
@@ -775,6 +881,7 @@ func (c *mcuProxyConnection) clearSubscribers() {
 			subscriber.NotifyClosed()
 		}
 	}(c.subscribers)
+	// Can't use clear(...) here as the map is processed by the goroutine above.
 	c.subscribers = make(map[string]*mcuProxySubscriber)
 
 	if c.closeScheduled.Load() || c.IsTemporary() {
@@ -786,7 +893,7 @@ func (c *mcuProxyConnection) clearCallbacks() {
 	c.mu.Lock()
 	defer c.mu.Unlock()
 
-	c.callbacks = make(map[string]func(*ProxyServerMessage))
+	clear(c.callbacks)
 }
 
 func (c *mcuProxyConnection) getCallback(id string) func(*ProxyServerMessage) {
@@ -806,11 +913,11 @@ func (c *mcuProxyConnection) processMessage(msg *ProxyServerMessage) {
 		switch msg.Type {
 		case "error":
 			if msg.Error.Code == "no_such_session" {
-				log.Printf("Session %s could not be resumed on %s, registering new", c.sessionId, c)
+				log.Printf("Session %s could not be resumed on %s, registering new", c.SessionId(), c)
 				c.clearPublishers()
 				c.clearSubscribers()
 				c.clearCallbacks()
-				c.sessionId = ""
+				c.sessionId.Store("")
 				if err := c.sendHello(); err != nil {
 					log.Printf("Could not send hello request to %s: %s", c, err)
 					c.scheduleReconnect()
@@ -821,8 +928,8 @@ func (c *mcuProxyConnection) processMessage(msg *ProxyServerMessage) {
 			log.Printf("Hello connection to %s failed with %+v, reconnecting", c, msg.Error)
 			c.scheduleReconnect()
 		case "hello":
-			resumed := c.sessionId == msg.Hello.SessionId
-			c.sessionId = msg.Hello.SessionId
+			resumed := c.SessionId() == msg.Hello.SessionId
+			c.sessionId.Store(msg.Hello.SessionId)
 			country := ""
 			if msg.Hello.Server != nil {
 				if country = msg.Hello.Server.Country; country != "" && !IsValidCountry(country) {
@@ -832,11 +939,11 @@ func (c *mcuProxyConnection) processMessage(msg *ProxyServerMessage) {
 			}
 			c.country.Store(country)
 			if resumed {
-				log.Printf("Resumed session %s on %s", c.sessionId, c)
+				log.Printf("Resumed session %s on %s", c.SessionId(), c)
 			} else if country != "" {
-				log.Printf("Received session %s from %s (in %s)", c.sessionId, c, country)
+				log.Printf("Received session %s from %s (in %s)", c.SessionId(), c, country)
 			} else {
-				log.Printf("Received session %s from %s", c.sessionId, c)
+				log.Printf("Received session %s from %s", c.SessionId(), c)
 			}
 			if c.trackClose.CompareAndSwap(false, true) {
 				statsConnectedProxyBackendsCurrent.WithLabelValues(c.Country()).Inc()
@@ -907,9 +1014,10 @@ func (c *mcuProxyConnection) processEvent(msg *ProxyServerMessage) {
 		return
 	case "update-load":
 		if proxyDebugMessages {
-			log.Printf("Load of %s now at %d", c, event.Load)
+			log.Printf("Load of %s now at %d (%s)", c, event.Load, event.Bandwidth)
 		}
 		c.load.Store(event.Load)
+		c.bandwidth.Store(event.Bandwidth)
 		statsProxyBackendLoadCurrent.WithLabelValues(c.url.String()).Set(float64(event.Load))
 		return
 	case "shutdown-scheduled":
@@ -944,8 +1052,14 @@ func (c *mcuProxyConnection) processBye(msg *ProxyServerMessage) {
 	bye := msg.Bye
 	switch bye.Reason {
 	case "session_resumed":
-		log.Printf("Session %s on %s was resumed by other client, resetting", c.sessionId, c)
-		c.sessionId = ""
+		log.Printf("Session %s on %s was resumed by other client, resetting", c.SessionId(), c)
+		c.sessionId.Store("")
+	case "session_expired":
+		log.Printf("Session %s expired on %s, resetting", c.SessionId(), c)
+		c.sessionId.Store("")
+	case "session_closed":
+		log.Printf("Session %s was closed on %s, resetting", c.SessionId(), c)
+		c.sessionId.Store("")
 	default:
 		log.Printf("Received bye with unsupported reason from %s %+v", c, bye)
 	}
@@ -960,17 +1074,10 @@ func (c *mcuProxyConnection) sendHello() error {
 			Version: "1.0",
 		},
 	}
-	if c.sessionId != "" {
-		msg.Hello.ResumeId = c.sessionId
+	if sessionId := c.SessionId(); sessionId != "" {
+		msg.Hello.ResumeId = sessionId
 	} else {
-		claims := &TokenClaims{
-			jwt.RegisteredClaims{
-				IssuedAt: jwt.NewNumericDate(time.Now()),
-				Issuer:   c.proxy.tokenId,
-			},
-		}
-		token := jwt.NewWithClaims(jwt.SigningMethodRS256, claims)
-		tokenString, err := token.SignedString(c.proxy.tokenKey)
+		tokenString, err := c.proxy.createToken("")
 		if err != nil {
 			return err
 		}
@@ -1039,15 +1146,17 @@ func (c *mcuProxyConnection) performSyncRequest(ctx context.Context, msg *ProxyC
 	}
 }
 
-func (c *mcuProxyConnection) newPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, bitrate int, mediaTypes MediaType) (McuPublisher, error) {
+func (c *mcuProxyConnection) newPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, settings NewPublisherSettings) (McuPublisher, error) {
 	msg := &ProxyClientMessage{
 		Type: "command",
 		Command: &CommandProxyClientMessage{
-			Type:       "create-publisher",
-			Sid:        sid,
-			StreamType: streamType,
-			Bitrate:    bitrate,
-			MediaTypes: mediaTypes,
+			Type:              "create-publisher",
+			Sid:               sid,
+			StreamType:        streamType,
+			PublisherSettings: &settings,
+			// Include for older version of the signaling proxy.
+			Bitrate:    settings.Bitrate,
+			MediaTypes: settings.MediaTypes,
 		},
 	}
 
@@ -1061,7 +1170,7 @@ func (c *mcuProxyConnection) newPublisher(ctx context.Context, listener McuListe
 
 	proxyId := response.Command.Id
 	log.Printf("Created %s publisher %s on %s for %s", streamType, proxyId, c, id)
-	publisher := newMcuProxyPublisher(id, sid, streamType, response.Command.Bitrate, mediaTypes, proxyId, c, listener)
+	publisher := newMcuProxyPublisher(id, sid, streamType, response.Command.Bitrate, settings, proxyId, c, listener)
 	c.publishersLock.Lock()
 	c.publishers[proxyId] = publisher
 	c.publisherIds[getStreamId(id, streamType)] = proxyId
@@ -1091,7 +1200,7 @@ func (c *mcuProxyConnection) newSubscriber(ctx context.Context, listener McuList
 
 	proxyId := response.Command.Id
 	log.Printf("Created %s subscriber %s on %s for %s", streamType, proxyId, c, publisherSessionId)
-	subscriber := newMcuProxySubscriber(publisherSessionId, response.Command.Sid, streamType, response.Command.Bitrate, proxyId, c, listener)
+	subscriber := newMcuProxySubscriber(publisherSessionId, response.Command.Sid, streamType, response.Command.Bitrate, proxyId, c, listener, nil)
 	c.subscribersLock.Lock()
 	c.subscribers[proxyId] = subscriber
 	c.subscribersLock.Unlock()
@@ -1100,6 +1209,81 @@ func (c *mcuProxyConnection) newSubscriber(ctx context.Context, listener McuList
 	return subscriber, nil
 }
 
+func (c *mcuProxyConnection) newRemoteSubscriber(ctx context.Context, listener McuListener, publisherId string, publisherSessionId string, streamType StreamType, publisherConn *mcuProxyConnection) (McuSubscriber, error) {
+	if c == publisherConn {
+		return c.newSubscriber(ctx, listener, publisherId, publisherSessionId, streamType)
+	}
+
+	remoteToken, err := c.proxy.createToken(publisherId)
+	if err != nil {
+		return nil, err
+	}
+
+	msg := &ProxyClientMessage{
+		Type: "command",
+		Command: &CommandProxyClientMessage{
+			Type:        "create-subscriber",
+			StreamType:  streamType,
+			PublisherId: publisherId,
+
+			RemoteUrl:   publisherConn.rawUrl,
+			RemoteToken: remoteToken,
+		},
+	}
+
+	response, err := c.performSyncRequest(ctx, msg)
+	if err != nil {
+		// TODO: Cancel request
+		return nil, err
+	} else if response.Type == "error" {
+		return nil, fmt.Errorf("Error creating remote %s subscriber for %s on %s (forwarded to %s): %+v", streamType, publisherSessionId, c, publisherConn, response.Error)
+	}
+
+	proxyId := response.Command.Id
+	log.Printf("Created remote %s subscriber %s on %s for %s (forwarded to %s)", streamType, proxyId, c, publisherSessionId, publisherConn)
+	subscriber := newMcuProxySubscriber(publisherSessionId, response.Command.Sid, streamType, response.Command.Bitrate, proxyId, c, listener, publisherConn)
+	c.subscribersLock.Lock()
+	c.subscribers[proxyId] = subscriber
+	c.subscribersLock.Unlock()
+	statsSubscribersCurrent.WithLabelValues(string(streamType)).Inc()
+	statsSubscribersTotal.WithLabelValues(string(streamType)).Inc()
+	return subscriber, nil
+}
+
+type mcuProxySettings struct {
+	mcuCommonSettings
+}
+
+func newMcuProxySettings(config *goconf.ConfigFile) (McuSettings, error) {
+	settings := &mcuProxySettings{}
+	if err := settings.load(config); err != nil {
+		return nil, err
+	}
+
+	return settings, nil
+}
+
+func (s *mcuProxySettings) load(config *goconf.ConfigFile) error {
+	if err := s.mcuCommonSettings.load(config); err != nil {
+		return err
+	}
+
+	proxyTimeoutSeconds, _ := config.GetInt("mcu", "proxytimeout")
+	if proxyTimeoutSeconds <= 0 {
+		proxyTimeoutSeconds = defaultProxyTimeoutSeconds
+	}
+	proxyTimeout := time.Duration(proxyTimeoutSeconds) * time.Second
+	log.Printf("Using a timeout of %s for proxy requests", proxyTimeout)
+	s.setTimeout(proxyTimeout)
+	return nil
+}
+
+func (s *mcuProxySettings) Reload(config *goconf.ConfigFile) {
+	if err := s.load(config); err != nil {
+		log.Printf("Error reloading proxy settings: %s", err)
+	}
+}
+
 type mcuProxy struct {
 	urlType  string
 	tokenId  string
@@ -1110,12 +1294,10 @@ type mcuProxy struct {
 	connections    []*mcuProxyConnection
 	connectionsMap map[string][]*mcuProxyConnection
 	connectionsMu  sync.RWMutex
-	proxyTimeout   time.Duration
 	connRequests   atomic.Int64
 	nextSort       atomic.Int64
 
-	maxStreamBitrate int
-	maxScreenBitrate int
+	settings McuSettings
 
 	mu         sync.RWMutex
 	publishers map[string]*mcuProxyConnection
@@ -1150,20 +1332,9 @@ func NewMcuProxy(config *goconf.ConfigFile, etcdClient *EtcdClient, rpcClients *
 		return nil, fmt.Errorf("Could not parse private key from %s: %s", tokenKeyFilename, err)
 	}
 
-	proxyTimeoutSeconds, _ := config.GetInt("mcu", "proxytimeout")
-	if proxyTimeoutSeconds <= 0 {
-		proxyTimeoutSeconds = defaultProxyTimeoutSeconds
-	}
-	proxyTimeout := time.Duration(proxyTimeoutSeconds) * time.Second
-	log.Printf("Using a timeout of %s for proxy requests", proxyTimeout)
-
-	maxStreamBitrate, _ := config.GetInt("mcu", "maxstreambitrate")
-	if maxStreamBitrate <= 0 {
-		maxStreamBitrate = defaultMaxStreamBitrate
-	}
-	maxScreenBitrate, _ := config.GetInt("mcu", "maxscreenbitrate")
-	if maxScreenBitrate <= 0 {
-		maxScreenBitrate = defaultMaxScreenBitrate
+	settings, err := newMcuProxySettings((config))
+	if err != nil {
+		return nil, err
 	}
 
 	mcu := &mcuProxy{
@@ -1173,13 +1344,10 @@ func NewMcuProxy(config *goconf.ConfigFile, etcdClient *EtcdClient, rpcClients *
 
 		dialer: &websocket.Dialer{
 			Proxy:            http.ProxyFromEnvironment,
-			HandshakeTimeout: proxyTimeout,
+			HandshakeTimeout: settings.Timeout(),
 		},
 		connectionsMap: make(map[string][]*mcuProxyConnection),
-		proxyTimeout:   proxyTimeout,
-
-		maxStreamBitrate: maxStreamBitrate,
-		maxScreenBitrate: maxScreenBitrate,
+		settings:       settings,
 
 		publishers: make(map[string]*mcuProxyConnection),
 
@@ -1254,10 +1422,7 @@ func (m *mcuProxy) loadContinentsMap(config *goconf.ConfigFile) error {
 	return nil
 }
 
-func (m *mcuProxy) Start() error {
-	log.Printf("Maximum bandwidth %d bits/sec per publishing stream", m.maxStreamBitrate)
-	log.Printf("Maximum bandwidth %d bits/sec per screensharing stream", m.maxScreenBitrate)
-
+func (m *mcuProxy) Start(ctx context.Context) error {
 	return m.config.Start()
 }
 
@@ -1274,6 +1439,48 @@ func (m *mcuProxy) Stop() {
 	m.config.Stop()
 }
 
+func (m *mcuProxy) createToken(subject string) (string, error) {
+	claims := &TokenClaims{
+		jwt.RegisteredClaims{
+			IssuedAt: jwt.NewNumericDate(time.Now()),
+			Issuer:   m.tokenId,
+			Subject:  subject,
+		},
+	}
+	token := jwt.NewWithClaims(jwt.SigningMethodRS256, claims)
+	tokenString, err := token.SignedString(m.tokenKey)
+	if err != nil {
+		return "", err
+	}
+
+	return tokenString, nil
+}
+
+func (m *mcuProxy) hasConnections() bool {
+	m.connectionsMu.RLock()
+	defer m.connectionsMu.RUnlock()
+	for _, conn := range m.connections {
+		if conn.IsConnected() {
+			return true
+		}
+	}
+	return false
+}
+
+func (m *mcuProxy) WaitForConnections(ctx context.Context) error {
+	ticker := time.NewTicker(10 * time.Millisecond)
+	defer ticker.Stop()
+
+	for !m.hasConnections() {
+		select {
+		case <-ctx.Done():
+			return ctx.Err()
+		case <-ticker.C:
+		}
+	}
+	return nil
+}
+
 func (m *mcuProxy) AddConnection(ignoreErrors bool, url string, ips ...net.IP) error {
 	m.connectionsMu.Lock()
 	defer m.connectionsMu.Unlock()
@@ -1373,6 +1580,12 @@ func (m *mcuProxy) KeepConnection(url string, ips ...net.IP) {
 }
 
 func (m *mcuProxy) Reload(config *goconf.ConfigFile) {
+	m.settings.Reload(config)
+
+	if m.settings.Timeout() != m.dialer.HandshakeTimeout {
+		m.dialer.HandshakeTimeout = m.settings.Timeout()
+	}
+
 	if err := m.loadContinentsMap(config); err != nil {
 		log.Printf("Error loading continents map: %s", err)
 	}
@@ -1565,28 +1778,30 @@ func (m *mcuProxy) removePublisher(publisher *mcuProxyPublisher) {
 	delete(m.publishers, getStreamId(publisher.id, publisher.StreamType()))
 }
 
-func (m *mcuProxy) NewPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, bitrate int, mediaTypes MediaType, initiator McuInitiator) (McuPublisher, error) {
-	connections := m.getSortedConnections(initiator)
+func (m *mcuProxy) createPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, settings NewPublisherSettings, initiator McuInitiator, connections []*mcuProxyConnection, isAllowed func(c *mcuProxyConnection) bool) McuPublisher {
+	var maxBitrate int
+	if streamType == StreamTypeScreen {
+		maxBitrate = int(m.settings.MaxScreenBitrate())
+	} else {
+		maxBitrate = int(m.settings.MaxStreamBitrate())
+	}
+
+	publisherSettings := settings
+	if publisherSettings.Bitrate <= 0 {
+		publisherSettings.Bitrate = maxBitrate
+	} else {
+		publisherSettings.Bitrate = min(publisherSettings.Bitrate, maxBitrate)
+	}
+
 	for _, conn := range connections {
-		if conn.IsShutdownScheduled() || conn.IsTemporary() {
+		if !isAllowed(conn) || conn.IsShutdownScheduled() || conn.IsTemporary() {
 			continue
 		}
 
-		subctx, cancel := context.WithTimeout(ctx, m.proxyTimeout)
+		subctx, cancel := context.WithTimeout(ctx, m.settings.Timeout())
 		defer cancel()
 
-		var maxBitrate int
-		if streamType == StreamTypeScreen {
-			maxBitrate = m.maxScreenBitrate
-		} else {
-			maxBitrate = m.maxStreamBitrate
-		}
-		if bitrate <= 0 {
-			bitrate = maxBitrate
-		} else {
-			bitrate = min(bitrate, maxBitrate)
-		}
-		publisher, err := conn.newPublisher(subctx, listener, id, sid, streamType, bitrate, mediaTypes)
+		publisher, err := conn.newPublisher(subctx, listener, id, sid, streamType, publisherSettings)
 		if err != nil {
 			log.Printf("Could not create %s publisher for %s on %s: %s", streamType, id, conn, err)
 			continue
@@ -1596,11 +1811,61 @@ func (m *mcuProxy) NewPublisher(ctx context.Context, listener McuListener, id st
 		m.publishers[getStreamId(id, streamType)] = conn
 		m.mu.Unlock()
 		m.publisherWaiters.Wakeup()
-		return publisher, nil
+		return publisher
 	}
 
-	statsProxyNobackendAvailableTotal.WithLabelValues(string(streamType)).Inc()
-	return nil, fmt.Errorf("No MCU connection available")
+	return nil
+}
+
+func (m *mcuProxy) NewPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, settings NewPublisherSettings, initiator McuInitiator) (McuPublisher, error) {
+	connections := m.getSortedConnections(initiator)
+	publisher := m.createPublisher(ctx, listener, id, sid, streamType, settings, initiator, connections, func(c *mcuProxyConnection) bool {
+		bw := c.Bandwidth()
+		return bw == nil || bw.AllowIncoming()
+	})
+	if publisher == nil {
+		// No proxy has available bandwidth, select one with the lowest currently used bandwidth.
+		connections2 := make([]*mcuProxyConnection, 0, len(connections))
+		for _, c := range connections {
+			if c.Bandwidth() != nil {
+				connections2 = append(connections2, c)
+			}
+		}
+		slices.SortFunc(connections2, func(a *mcuProxyConnection, b *mcuProxyConnection) int {
+			var incoming_a *float64
+			if bw := a.Bandwidth(); bw != nil {
+				incoming_a = bw.Incoming
+			}
+
+			var incoming_b *float64
+			if bw := b.Bandwidth(); bw != nil {
+				incoming_b = bw.Incoming
+			}
+
+			if incoming_a == nil && incoming_b == nil {
+				return 0
+			} else if incoming_a == nil && incoming_b != nil {
+				return -1
+			} else if incoming_a != nil && incoming_b == nil {
+				return -1
+			} else if *incoming_a < *incoming_b {
+				return -1
+			} else if *incoming_a > *incoming_b {
+				return 1
+			}
+			return 0
+		})
+		publisher = m.createPublisher(ctx, listener, id, sid, streamType, settings, initiator, connections2, func(c *mcuProxyConnection) bool {
+			return true
+		})
+	}
+
+	if publisher == nil {
+		statsProxyNobackendAvailableTotal.WithLabelValues(string(streamType)).Inc()
+		return nil, fmt.Errorf("No MCU connection available")
+	}
+
+	return publisher, nil
 }
 
 func (m *mcuProxy) getPublisherConnection(publisher string, streamType StreamType) *mcuProxyConnection {
@@ -1641,7 +1906,38 @@ func (m *mcuProxy) waitForPublisherConnection(ctx context.Context, publisher str
 	}
 }
 
-func (m *mcuProxy) NewSubscriber(ctx context.Context, listener McuListener, publisher string, streamType StreamType) (McuSubscriber, error) {
+type proxyPublisherInfo struct {
+	id   string
+	conn *mcuProxyConnection
+	err  error
+}
+
+func (m *mcuProxy) createSubscriber(ctx context.Context, listener McuListener, id string, publisher string, streamType StreamType, publisherConn *mcuProxyConnection, connections []*mcuProxyConnection, isAllowed func(c *mcuProxyConnection) bool) McuSubscriber {
+	for _, conn := range connections {
+		if !isAllowed(conn) || conn.IsShutdownScheduled() || conn.IsTemporary() {
+			continue
+		}
+
+		var subscriber McuSubscriber
+		var err error
+		if conn == publisherConn {
+			subscriber, err = conn.newSubscriber(ctx, listener, id, publisher, streamType)
+		} else {
+			subscriber, err = conn.newRemoteSubscriber(ctx, listener, id, publisher, streamType, publisherConn)
+		}
+		if err != nil {
+			log.Printf("Could not create subscriber for %s publisher %s on %s: %s", streamType, publisher, conn, err)
+			continue
+		}
+
+		return subscriber
+	}
+
+	return nil
+}
+
+func (m *mcuProxy) NewSubscriber(ctx context.Context, listener McuListener, publisher string, streamType StreamType, initiator McuInitiator) (McuSubscriber, error) {
+	var publisherInfo *proxyPublisherInfo
 	if conn := m.getPublisherConnection(publisher, streamType); conn != nil {
 		// Fast common path: publisher is available locally.
 		conn.publishersLock.Lock()
@@ -1651,113 +1947,190 @@ func (m *mcuProxy) NewSubscriber(ctx context.Context, listener McuListener, publ
 			return nil, fmt.Errorf("Unknown publisher %s", publisher)
 		}
 
-		return conn.newSubscriber(ctx, listener, id, publisher, streamType)
-	}
-
-	log.Printf("No %s publisher %s found yet, deferring", streamType, publisher)
-	ch := make(chan McuSubscriber)
-	getctx, cancel := context.WithCancel(ctx)
-	defer cancel()
-
-	// Wait for publisher to be created locally.
-	go func() {
-		if conn := m.waitForPublisherConnection(getctx, publisher, streamType); conn != nil {
-			cancel() // Cancel pending RPC calls.
-
-			conn.publishersLock.Lock()
-			id, found := conn.publisherIds[getStreamId(publisher, streamType)]
-			conn.publishersLock.Unlock()
-			if !found {
-				log.Printf("Unknown id for local %s publisher %s", streamType, publisher)
-				return
-			}
-
-			subscriber, err := conn.newSubscriber(ctx, listener, id, publisher, streamType)
-			if subscriber != nil {
-				ch <- subscriber
-			} else if err != nil {
-				log.Printf("Error creating local subscriber for %s publisher %s: %s", streamType, publisher, err)
-			}
+		publisherInfo = &proxyPublisherInfo{
+			id:   id,
+			conn: conn,
 		}
-	}()
+	} else {
+		log.Printf("No %s publisher %s found yet, deferring", streamType, publisher)
+		ch := make(chan *proxyPublisherInfo, 1)
+		getctx, cancel := context.WithCancel(ctx)
+		defer cancel()
 
-	// Wait for publisher to be created on one of the other servers in the cluster.
-	if clients := m.rpcClients.GetClients(); len(clients) > 0 {
-		for _, client := range clients {
-			go func(client *GrpcClient) {
-				id, url, ip, err := client.GetPublisherId(getctx, publisher, streamType)
-				if errors.Is(err, context.Canceled) {
-					return
-				} else if err != nil {
-					log.Printf("Error getting %s publisher id %s from %s: %s", streamType, publisher, client.Target(), err)
-					return
-				} else if id == "" {
-					// Publisher not found on other server
-					return
-				}
+		var wg sync.WaitGroup
 
+		// Wait for publisher to be created locally.
+		wg.Add(1)
+		go func() {
+			defer wg.Done()
+			if conn := m.waitForPublisherConnection(getctx, publisher, streamType); conn != nil {
 				cancel() // Cancel pending RPC calls.
-				log.Printf("Found publisher id %s through %s on proxy %s", id, client.Target(), url)
 
-				m.connectionsMu.RLock()
-				connections := m.connections
-				m.connectionsMu.RUnlock()
-				var publisherConn *mcuProxyConnection
-				for _, conn := range connections {
-					if conn.rawUrl != url || !ip.Equal(conn.ip) {
-						continue
+				conn.publishersLock.Lock()
+				id, found := conn.publisherIds[getStreamId(publisher, streamType)]
+				conn.publishersLock.Unlock()
+				if !found {
+					ch <- &proxyPublisherInfo{
+						err: fmt.Errorf("Unknown id for local %s publisher %s", streamType, publisher),
 					}
-
-					// Simple case, signaling server has a connection to the same endpoint
-					publisherConn = conn
-					break
-				}
-
-				if publisherConn == nil {
-					publisherConn, err = newMcuProxyConnection(m, url, ip)
-					if err != nil {
-						log.Printf("Could not create temporary connection to %s for %s publisher %s: %s", url, streamType, publisher, err)
-						return
-					}
-
-					publisherConn.setTemporary()
-					publisherConn.start()
-					if err := publisherConn.waitUntilConnected(ctx); err != nil {
-						log.Printf("Could not establish new connection to %s: %s", publisherConn, err)
-						publisherConn.closeIfEmpty()
-						return
-					}
-
-					m.connectionsMu.Lock()
-					m.connections = append(m.connections, publisherConn)
-					conns, found := m.connectionsMap[url]
-					if found {
-						conns = append(conns, publisherConn)
-					} else {
-						conns = []*mcuProxyConnection{publisherConn}
-					}
-					m.connectionsMap[url] = conns
-					m.connectionsMu.Unlock()
-				}
-
-				subscriber, err := publisherConn.newSubscriber(ctx, listener, id, publisher, streamType)
-				if err != nil {
-					if publisherConn.IsTemporary() {
-						publisherConn.closeIfEmpty()
-					}
-					log.Printf("Could not create subscriber for %s publisher %s: %s", streamType, publisher, err)
 					return
 				}
 
-				ch <- subscriber
-			}(client)
+				ch <- &proxyPublisherInfo{
+					id:   id,
+					conn: conn,
+				}
+			}
+		}()
+
+		// Wait for publisher to be created on one of the other servers in the cluster.
+		if clients := m.rpcClients.GetClients(); len(clients) > 0 {
+			for _, client := range clients {
+				wg.Add(1)
+				go func(client *GrpcClient) {
+					defer wg.Done()
+					id, url, ip, err := client.GetPublisherId(getctx, publisher, streamType)
+					if errors.Is(err, context.Canceled) {
+						return
+					} else if err != nil {
+						log.Printf("Error getting %s publisher id %s from %s: %s", streamType, publisher, client.Target(), err)
+						return
+					} else if id == "" {
+						// Publisher not found on other server
+						return
+					}
+
+					cancel() // Cancel pending RPC calls.
+					log.Printf("Found publisher id %s through %s on proxy %s", id, client.Target(), url)
+
+					m.connectionsMu.RLock()
+					connections := m.connections
+					m.connectionsMu.RUnlock()
+					var publisherConn *mcuProxyConnection
+					for _, conn := range connections {
+						if conn.rawUrl != url || !ip.Equal(conn.ip) {
+							continue
+						}
+
+						// Simple case, signaling server has a connection to the same endpoint
+						publisherConn = conn
+						break
+					}
+
+					if publisherConn == nil {
+						publisherConn, err = newMcuProxyConnection(m, url, ip)
+						if err != nil {
+							log.Printf("Could not create temporary connection to %s for %s publisher %s: %s", url, streamType, publisher, err)
+							return
+						}
+
+						publisherConn.setTemporary()
+						publisherConn.start()
+						if err := publisherConn.waitUntilConnected(ctx); err != nil {
+							log.Printf("Could not establish new connection to %s: %s", publisherConn, err)
+							publisherConn.closeIfEmpty()
+							return
+						}
+
+						m.connectionsMu.Lock()
+						m.connections = append(m.connections, publisherConn)
+						conns, found := m.connectionsMap[url]
+						if found {
+							conns = append(conns, publisherConn)
+						} else {
+							conns = []*mcuProxyConnection{publisherConn}
+						}
+						m.connectionsMap[url] = conns
+						m.connectionsMu.Unlock()
+					}
+
+					ch <- &proxyPublisherInfo{
+						id:   id,
+						conn: publisherConn,
+					}
+				}(client)
+			}
+		}
+
+		wg.Wait()
+		select {
+		case ch <- &proxyPublisherInfo{
+			err: fmt.Errorf("No %s publisher %s found", streamType, publisher),
+		}:
+		default:
+		}
+
+		select {
+		case info := <-ch:
+			publisherInfo = info
+		case <-ctx.Done():
+			return nil, fmt.Errorf("No %s publisher %s found", streamType, publisher)
 		}
 	}
 
-	select {
-	case subscriber := <-ch:
-		return subscriber, nil
-	case <-ctx.Done():
-		return nil, fmt.Errorf("No %s publisher %s found", streamType, publisher)
+	if publisherInfo.err != nil {
+		return nil, publisherInfo.err
 	}
+
+	bw := publisherInfo.conn.Bandwidth()
+	allowOutgoing := bw == nil || bw.AllowOutgoing()
+	if !allowOutgoing || !publisherInfo.conn.IsSameCountry(initiator) {
+		connections := m.getSortedConnections(initiator)
+		if !allowOutgoing || len(connections) > 0 && !connections[0].IsSameCountry(publisherInfo.conn) {
+			// Connect to remote publisher through "closer" gateway.
+			subscriber := m.createSubscriber(ctx, listener, publisherInfo.id, publisher, streamType, publisherInfo.conn, connections, func(c *mcuProxyConnection) bool {
+				bw := c.Bandwidth()
+				return bw == nil || bw.AllowOutgoing()
+			})
+			if subscriber == nil {
+				connections2 := make([]*mcuProxyConnection, 0, len(connections))
+				for _, c := range connections {
+					if c.Bandwidth() != nil {
+						connections2 = append(connections2, c)
+					}
+				}
+				slices.SortFunc(connections2, func(a *mcuProxyConnection, b *mcuProxyConnection) int {
+					var outgoing_a *float64
+					if bw := a.Bandwidth(); bw != nil {
+						outgoing_a = bw.Outgoing
+					}
+
+					var outgoing_b *float64
+					if bw := b.Bandwidth(); bw != nil {
+						outgoing_b = bw.Outgoing
+					}
+
+					if outgoing_a == nil && outgoing_b == nil {
+						return 0
+					} else if outgoing_a == nil && outgoing_b != nil {
+						return -1
+					} else if outgoing_a != nil && outgoing_b == nil {
+						return -1
+					} else if *outgoing_a < *outgoing_b {
+						return -1
+					} else if *outgoing_a > *outgoing_b {
+						return 1
+					}
+					return 0
+				})
+				subscriber = m.createSubscriber(ctx, listener, publisherInfo.id, publisher, streamType, publisherInfo.conn, connections2, func(c *mcuProxyConnection) bool {
+					return true
+				})
+			}
+			if subscriber != nil {
+				return subscriber, nil
+			}
+		}
+	}
+
+	subscriber, err := publisherInfo.conn.newSubscriber(ctx, listener, publisherInfo.id, publisher, streamType)
+	if err != nil {
+		if publisherInfo.conn.IsTemporary() {
+			publisherInfo.conn.closeIfEmpty()
+		}
+		log.Printf("Could not create subscriber for %s publisher %s on %s: %s", streamType, publisher, publisherInfo.conn, err)
+		return nil, err
+	}
+
+	return subscriber, nil
 }
