@@ -23,6 +23,7 @@ package signaling
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"log"
 	"sync"
@@ -49,7 +50,7 @@ func NewTestMCU() (*TestMCU, error) {
 	}, nil
 }
 
-func (m *TestMCU) Start() error {
+func (m *TestMCU) Start(ctx context.Context) error {
 	return nil
 }
 
@@ -69,17 +70,17 @@ func (m *TestMCU) GetStats() interface{} {
 	return nil
 }
 
-func (m *TestMCU) NewPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, bitrate int, mediaTypes MediaType, initiator McuInitiator) (McuPublisher, error) {
+func (m *TestMCU) NewPublisher(ctx context.Context, listener McuListener, id string, sid string, streamType StreamType, settings NewPublisherSettings, initiator McuInitiator) (McuPublisher, error) {
 	var maxBitrate int
 	if streamType == StreamTypeScreen {
 		maxBitrate = TestMaxBitrateScreen
 	} else {
 		maxBitrate = TestMaxBitrateVideo
 	}
-	if bitrate <= 0 {
-		bitrate = maxBitrate
-	} else if bitrate > maxBitrate {
-		bitrate = maxBitrate
+	publisherSettings := settings
+	bitrate := publisherSettings.Bitrate
+	if bitrate <= 0 || bitrate > maxBitrate {
+		publisherSettings.Bitrate = maxBitrate
 	}
 	pub := &TestMCUPublisher{
 		TestMCUClient: TestMCUClient{
@@ -88,8 +89,7 @@ func (m *TestMCU) NewPublisher(ctx context.Context, listener McuListener, id str
 			streamType: streamType,
 		},
 
-		mediaTypes: mediaTypes,
-		bitrate:    bitrate,
+		settings: publisherSettings,
 	}
 
 	m.mu.Lock()
@@ -117,7 +117,7 @@ func (m *TestMCU) GetPublisher(id string) *TestMCUPublisher {
 	return m.publishers[id]
 }
 
-func (m *TestMCU) NewSubscriber(ctx context.Context, listener McuListener, publisher string, streamType StreamType) (McuSubscriber, error) {
+func (m *TestMCU) NewSubscriber(ctx context.Context, listener McuListener, publisher string, streamType StreamType, initiator McuInitiator) (McuSubscriber, error) {
 	m.mu.Lock()
 	defer m.mu.Unlock()
 
@@ -175,18 +175,17 @@ func (c *TestMCUClient) isClosed() bool {
 type TestMCUPublisher struct {
 	TestMCUClient
 
-	mediaTypes MediaType
-	bitrate    int
+	settings NewPublisherSettings
 
 	sdp string
 }
 
 func (p *TestMCUPublisher) HasMedia(mt MediaType) bool {
-	return (p.mediaTypes & mt) == mt
+	return (p.settings.MediaTypes & mt) == mt
 }
 
 func (p *TestMCUPublisher) SetMedia(mt MediaType) {
-	p.mediaTypes = mt
+	p.settings.MediaTypes = mt
 }
 
 func (p *TestMCUPublisher) SendMessage(ctx context.Context, message *MessageClientMessage, data *MessageClientMessageData, callback func(error, map[string]interface{})) {
@@ -222,6 +221,18 @@ func (p *TestMCUPublisher) SendMessage(ctx context.Context, message *MessageClie
 	}()
 }
 
+func (p *TestMCUPublisher) GetStreams(ctx context.Context) ([]PublisherStream, error) {
+	return nil, errors.New("not implemented")
+}
+
+func (p *TestMCUPublisher) PublishRemote(ctx context.Context, remoteId string, hostname string, port int, rtcpPort int) error {
+	return errors.New("remote publishing not supported")
+}
+
+func (p *TestMCUPublisher) UnpublishRemote(ctx context.Context, remoteId string, hostname string, port int, rtcpPort int) error {
+	return errors.New("remote publishing not supported")
+}
+
 type TestMCUSubscriber struct {
 	TestMCUClient
 
@@ -253,6 +264,8 @@ func (s *TestMCUSubscriber) SendMessage(ctx context.Context, message *MessageCli
 				"type": "offer",
 				"sdp":  sdp,
 			})
+		case "answer":
+			callback(nil, nil)
 		default:
 			callback(fmt.Errorf("Message type %s is not implemented", data.Type), nil)
 		}
