@@ -24,7 +24,10 @@ package main
 import (
 	"context"
 	"crypto/rand"
+	"crypto/rsa"
+	"crypto/tls"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
 	"log"
@@ -40,12 +43,13 @@ import (
 	"time"
 
 	"github.com/dlintw/goconf"
-	"github.com/golang-jwt/jwt/v4"
+	"github.com/golang-jwt/jwt/v5"
 	"github.com/google/uuid"
 	"github.com/gorilla/mux"
-	"github.com/gorilla/securecookie"
 	"github.com/gorilla/websocket"
+	"github.com/notedit/janus-go"
 	"github.com/prometheus/client_golang/prometheus/promhttp"
+	"google.golang.org/protobuf/types/known/timestamppb"
 
 	signaling "github.com/strukturag/nextcloud-spreed-signaling"
 )
@@ -58,11 +62,27 @@ const (
 	initialMcuRetry = time.Second
 	maxMcuRetry     = time.Second * 16
 
+	// MCU requests will be cancelled if they take too long.
+	defaultMcuTimeoutSeconds = 10
+
 	updateLoadInterval     = time.Second
 	expireSessionsInterval = 10 * time.Second
 
 	// Maximum age a token may have to prevent reuse of old tokens.
 	maxTokenAge = 5 * time.Minute
+
+	// Allow time differences of up to one minute between server and proxy.
+	tokenLeeway = time.Minute
+
+	remotePublisherTimeout = 5 * time.Second
+
+	ProxyFeatureRemoteStreams = "remote-streams"
+)
+
+var (
+	defaultProxyFeatures = []string{
+		ProxyFeatureRemoteStreams,
+	}
 )
 
 type ContextKey string
@@ -70,44 +90,119 @@ type ContextKey string
 var (
 	ContextKeySession = ContextKey("session")
 
-	TimeoutCreatingPublisher  = signaling.NewError("timeout", "Timeout creating publisher.")
-	TimeoutCreatingSubscriber = signaling.NewError("timeout", "Timeout creating subscriber.")
-	TokenAuthFailed           = signaling.NewError("auth_failed", "The token could not be authenticated.")
-	TokenExpired              = signaling.NewError("token_expired", "The token is expired.")
-	TokenNotValidYet          = signaling.NewError("token_not_valid_yet", "The token is not valid yet.")
-	UnknownClient             = signaling.NewError("unknown_client", "Unknown client id given.")
-	UnsupportedCommand        = signaling.NewError("bad_request", "Unsupported command received.")
-	UnsupportedMessage        = signaling.NewError("bad_request", "Unsupported message received.")
-	UnsupportedPayload        = signaling.NewError("unsupported_payload", "Unsupported payload type.")
-	ShutdownScheduled         = signaling.NewError("shutdown_scheduled", "The server is scheduled to shutdown.")
+	TimeoutCreatingPublisher      = signaling.NewError("timeout", "Timeout creating publisher.")
+	TimeoutCreatingSubscriber     = signaling.NewError("timeout", "Timeout creating subscriber.")
+	TokenAuthFailed               = signaling.NewError("auth_failed", "The token could not be authenticated.")
+	TokenExpired                  = signaling.NewError("token_expired", "The token is expired.")
+	TokenNotValidYet              = signaling.NewError("token_not_valid_yet", "The token is not valid yet.")
+	UnknownClient                 = signaling.NewError("unknown_client", "Unknown client id given.")
+	UnsupportedCommand            = signaling.NewError("bad_request", "Unsupported command received.")
+	UnsupportedMessage            = signaling.NewError("bad_request", "Unsupported message received.")
+	UnsupportedPayload            = signaling.NewError("unsupported_payload", "Unsupported payload type.")
+	ShutdownScheduled             = signaling.NewError("shutdown_scheduled", "The server is scheduled to shutdown.")
+	RemoteSubscribersNotSupported = signaling.NewError("unsupported_subscriber", "Remote subscribers are not supported.")
 )
 
 type ProxyServer struct {
 	version        string
 	country        string
 	welcomeMessage string
+	welcomeMsg     *signaling.WelcomeServerMessage
+	config         *goconf.ConfigFile
+	mcuTimeout     time.Duration
 
 	url     string
 	mcu     signaling.Mcu
 	stopped atomic.Bool
 	load    atomic.Int64
 
+	maxIncoming     atomic.Int64
+	currentIncoming atomic.Int64
+	maxOutgoing     atomic.Int64
+	currentOutgoing atomic.Int64
+
 	shutdownChannel   chan struct{}
 	shutdownScheduled atomic.Bool
 
 	upgrader websocket.Upgrader
 
 	tokens          ProxyTokens
-	statsAllowedIps *signaling.AllowedIps
+	statsAllowedIps atomic.Pointer[signaling.AllowedIps]
+	trustedProxies  atomic.Pointer[signaling.AllowedIps]
 
 	sid          atomic.Uint64
-	cookie       *securecookie.SecureCookie
+	cookie       *signaling.SessionIdCodec
 	sessions     map[uint64]*ProxySession
 	sessionsLock sync.RWMutex
 
 	clients     map[string]signaling.McuClient
 	clientIds   map[string]string
 	clientsLock sync.RWMutex
+
+	tokenId               string
+	tokenKey              *rsa.PrivateKey
+	remoteTlsConfig       *tls.Config
+	remoteHostname        string
+	remoteConnections     map[string]*RemoteConnection
+	remoteConnectionsLock sync.Mutex
+}
+
+func IsPublicIP(IP net.IP) bool {
+	if IP.IsLoopback() || IP.IsLinkLocalMulticast() || IP.IsLinkLocalUnicast() {
+		return false
+	}
+	if ip4 := IP.To4(); ip4 != nil {
+		switch {
+		case ip4[0] == 10:
+			return false
+		case ip4[0] == 172 && ip4[1] >= 16 && ip4[1] <= 31:
+			return false
+		case ip4[0] == 192 && ip4[1] == 168:
+			return false
+		default:
+			return true
+		}
+	}
+	return false
+}
+
+func GetLocalIP() (string, error) {
+	addrs, err := net.InterfaceAddrs()
+	if err != nil {
+		return "", err
+	}
+
+	for _, address := range addrs {
+		if ipnet, ok := address.(*net.IPNet); ok && IsPublicIP(ipnet.IP) {
+			if ipnet.IP.To4() != nil {
+				return ipnet.IP.String(), nil
+			}
+		}
+	}
+	return "", nil
+}
+
+func getTargetBandwidths(config *goconf.ConfigFile) (int, int) {
+	maxIncoming, _ := config.GetInt("bandwidth", "incoming")
+	if maxIncoming < 0 {
+		maxIncoming = 0
+	}
+	if maxIncoming > 0 {
+		log.Printf("Target bandwidth for incoming streams: %d MBit/s", maxIncoming)
+	} else {
+		log.Printf("Target bandwidth for incoming streams: unlimited")
+	}
+	maxOutgoing, _ := config.GetInt("bandwidth", "outgoing")
+	if maxOutgoing < 0 {
+		maxOutgoing = 0
+	}
+	if maxIncoming > 0 {
+		log.Printf("Target bandwidth for outgoing streams: %d MBit/s", maxOutgoing)
+	} else {
+		log.Printf("Target bandwidth for outgoing streams: unlimited")
+	}
+
+	return maxIncoming, maxOutgoing
 }
 
 func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*ProxyServer, error) {
@@ -153,6 +248,19 @@ func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*
 		statsAllowedIps = signaling.DefaultAllowedIps()
 	}
 
+	trustedProxies, _ := config.GetString("app", "trustedproxies")
+	trustedProxiesIps, err := signaling.ParseAllowedIps(trustedProxies)
+	if err != nil {
+		return nil, err
+	}
+
+	if !trustedProxiesIps.Empty() {
+		log.Printf("Trusted proxies: %s", trustedProxiesIps)
+	} else {
+		trustedProxiesIps = signaling.DefaultTrustedProxies
+		log.Printf("No trusted proxies configured, only allowing for %s", trustedProxiesIps)
+	}
+
 	country, _ := config.GetString("app", "country")
 	country = strings.ToUpper(country)
 	if signaling.IsValidCountry(country) {
@@ -173,10 +281,68 @@ func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*
 		return nil, err
 	}
 
+	tokenId, _ := config.GetString("app", "token_id")
+	var tokenKey *rsa.PrivateKey
+	var remoteHostname string
+	var remoteTlsConfig *tls.Config
+	if tokenId != "" {
+		tokenKeyFilename, _ := config.GetString("app", "token_key")
+		if tokenKeyFilename == "" {
+			return nil, fmt.Errorf("No token key configured")
+		}
+		tokenKeyData, err := os.ReadFile(tokenKeyFilename)
+		if err != nil {
+			return nil, fmt.Errorf("Could not read private key from %s: %s", tokenKeyFilename, err)
+		}
+		tokenKey, err = jwt.ParseRSAPrivateKeyFromPEM(tokenKeyData)
+		if err != nil {
+			return nil, fmt.Errorf("Could not parse private key from %s: %s", tokenKeyFilename, err)
+		}
+		log.Printf("Using \"%s\" as token id for remote streams", tokenId)
+
+		remoteHostname, _ = config.GetString("app", "hostname")
+		if remoteHostname == "" {
+			remoteHostname, err = GetLocalIP()
+			if err != nil {
+				return nil, fmt.Errorf("could not get local ip: %w", err)
+			}
+		}
+		if remoteHostname == "" {
+			log.Printf("WARNING: Could not determine hostname for remote streams, will be disabled. Please configure manually.")
+		} else {
+			log.Printf("Using \"%s\" as hostname for remote streams", remoteHostname)
+		}
+
+		skipverify, _ := config.GetBool("backend", "skipverify")
+		if skipverify {
+			log.Println("WARNING: Remote stream requests verification is disabled!")
+			remoteTlsConfig = &tls.Config{
+				InsecureSkipVerify: skipverify,
+			}
+		}
+	} else {
+		log.Printf("No token id configured, remote streams will be disabled")
+	}
+
+	maxIncoming, maxOutgoing := getTargetBandwidths(config)
+
+	mcuTimeoutSeconds, _ := config.GetInt("mcu", "timeout")
+	if mcuTimeoutSeconds <= 0 {
+		mcuTimeoutSeconds = defaultMcuTimeoutSeconds
+	}
+	mcuTimeout := time.Duration(mcuTimeoutSeconds) * time.Second
+
 	result := &ProxyServer{
 		version:        version,
 		country:        country,
 		welcomeMessage: string(welcomeMessage) + "\n",
+		welcomeMsg: &signaling.WelcomeServerMessage{
+			Version:  version,
+			Country:  country,
+			Features: defaultProxyFeatures,
+		},
+		config:     config,
+		mcuTimeout: mcuTimeout,
 
 		shutdownChannel: make(chan struct{}),
 
@@ -185,16 +351,25 @@ func NewProxyServer(r *mux.Router, version string, config *goconf.ConfigFile) (*
 			WriteBufferSize: websocketWriteBufferSize,
 		},
 
-		tokens:          tokens,
-		statsAllowedIps: statsAllowedIps,
+		tokens: tokens,
 
-		cookie:   securecookie.New(hashKey, blockKey).MaxAge(0),
+		cookie:   signaling.NewSessionIdCodec(hashKey, blockKey),
 		sessions: make(map[uint64]*ProxySession),
 
 		clients:   make(map[string]signaling.McuClient),
 		clientIds: make(map[string]string),
+
+		tokenId:           tokenId,
+		tokenKey:          tokenKey,
+		remoteTlsConfig:   remoteTlsConfig,
+		remoteHostname:    remoteHostname,
+		remoteConnections: make(map[string]*RemoteConnection),
 	}
 
+	result.maxIncoming.Store(int64(maxIncoming) * 1024 * 1024)
+	result.maxOutgoing.Store(int64(maxOutgoing) * 1024 * 1024)
+	result.statsAllowedIps.Store(statsAllowedIps)
+	result.trustedProxies.Store(trustedProxiesIps)
 	result.upgrader.CheckOrigin = result.checkOrigin
 
 	if debug, _ := config.GetBool("app", "debug"); debug {
@@ -223,7 +398,7 @@ func (s *ProxyServer) checkOrigin(r *http.Request) bool {
 }
 
 func (s *ProxyServer) Start(config *goconf.ConfigFile) error {
-	s.url, _ = config.GetString("mcu", "url")
+	s.url, _ = signaling.GetStringOptionWithEnv(config, "mcu", "url")
 	if s.url == "" {
 		return fmt.Errorf("No MCU server url configured")
 	}
@@ -245,7 +420,7 @@ func (s *ProxyServer) Start(config *goconf.ConfigFile) error {
 	for {
 		switch mcuType {
 		case signaling.McuTypeJanus:
-			mcu, err = signaling.NewMcuJanus(s.url, config)
+			mcu, err = signaling.NewMcuJanus(ctx, s.url, config)
 			if err == nil {
 				signaling.RegisterJanusMcuStats()
 			}
@@ -255,7 +430,7 @@ func (s *ProxyServer) Start(config *goconf.ConfigFile) error {
 		if err == nil {
 			mcu.SetOnConnected(s.onMcuConnected)
 			mcu.SetOnDisconnected(s.onMcuDisconnected)
-			err = mcu.Start()
+			err = mcu.Start(ctx)
 			if err != nil {
 				log.Printf("Could not create %s MCU at %s: %s", mcuType, s.url, err)
 			}
@@ -298,18 +473,7 @@ loop:
 	}
 }
 
-func (s *ProxyServer) updateLoad() {
-	load := s.GetClientsLoad()
-	if load == s.load.Load() {
-		return
-	}
-
-	s.load.Store(load)
-	if s.shutdownScheduled.Load() {
-		// Server is scheduled to shutdown, no need to update clients with current load.
-		return
-	}
-
+func (s *ProxyServer) newLoadEvent(load int64, incoming int64, outgoing int64) *signaling.ProxyServerMessage {
 	msg := &signaling.ProxyServerMessage{
 		Type: "event",
 		Event: &signaling.EventProxyServerMessage{
@@ -317,7 +481,41 @@ func (s *ProxyServer) updateLoad() {
 			Load: load,
 		},
 	}
+	maxIncoming := s.maxIncoming.Load()
+	maxOutgoing := s.maxOutgoing.Load()
+	if maxIncoming > 0 || maxOutgoing > 0 {
+		msg.Event.Bandwidth = &signaling.EventProxyServerBandwidth{}
+		if maxIncoming > 0 {
+			value := float64(incoming) / float64(maxIncoming) * 100
+			msg.Event.Bandwidth.Incoming = &value
+		}
+		if maxOutgoing > 0 {
+			value := float64(outgoing) / float64(maxOutgoing) * 100
+			msg.Event.Bandwidth.Outgoing = &value
+		}
+	}
+	return msg
+}
 
+func (s *ProxyServer) updateLoad() {
+	load, incoming, outgoing := s.GetClientsLoad()
+	oldLoad := s.load.Swap(load)
+	oldIncoming := s.currentIncoming.Swap(incoming)
+	oldOutgoing := s.currentOutgoing.Swap(outgoing)
+	if oldLoad == load && oldIncoming == incoming && oldOutgoing == outgoing {
+		return
+	}
+
+	s.sendLoadToAll(load, incoming, outgoing)
+}
+
+func (s *ProxyServer) sendLoadToAll(load int64, incoming int64, outgoing int64) {
+	if s.shutdownScheduled.Load() {
+		// Server is scheduled to shutdown, no need to update clients with current load.
+		return
+	}
+
+	msg := s.newLoadEvent(load, incoming, outgoing)
 	s.IterateSessions(func(session *ProxySession) {
 		session.sendMessage(msg)
 	})
@@ -388,7 +586,42 @@ func (s *ProxyServer) ScheduleShutdown() {
 }
 
 func (s *ProxyServer) Reload(config *goconf.ConfigFile) {
+	statsAllowed, _ := config.GetString("stats", "allowed_ips")
+	if statsAllowedIps, err := signaling.ParseAllowedIps(statsAllowed); err == nil {
+		if !statsAllowedIps.Empty() {
+			log.Printf("Only allowing access to the stats endpoint from %s", statsAllowed)
+		} else {
+			log.Printf("No IPs configured for the stats endpoint, only allowing access from 127.0.0.1")
+			statsAllowedIps = signaling.DefaultAllowedIps()
+		}
+		s.statsAllowedIps.Store(statsAllowedIps)
+	} else {
+		log.Printf("Error parsing allowed stats ips from \"%s\": %s", statsAllowedIps, err)
+	}
+
+	trustedProxies, _ := config.GetString("app", "trustedproxies")
+	if trustedProxiesIps, err := signaling.ParseAllowedIps(trustedProxies); err == nil {
+		if !trustedProxiesIps.Empty() {
+			log.Printf("Trusted proxies: %s", trustedProxiesIps)
+		} else {
+			trustedProxiesIps = signaling.DefaultTrustedProxies
+			log.Printf("No trusted proxies configured, only allowing for %s", trustedProxiesIps)
+		}
+		s.trustedProxies.Store(trustedProxiesIps)
+	} else {
+		log.Printf("Error parsing trusted proxies from \"%s\": %s", trustedProxies, err)
+	}
+
+	maxIncoming, maxOutgoing := getTargetBandwidths(config)
+	oldIncoming := s.maxIncoming.Swap(int64(maxIncoming))
+	oldOutgoing := s.maxOutgoing.Swap(int64(maxOutgoing))
+	if oldIncoming != int64(maxIncoming) || oldOutgoing != int64(maxOutgoing) {
+		// Notify sessions about updated load / bandwidth usage.
+		go s.sendLoadToAll(s.load.Load(), s.currentIncoming.Load(), s.currentOutgoing.Load())
+	}
+
 	s.tokens.Reload(config)
+	s.mcu.Reload(config)
 }
 
 func (s *ProxyServer) setCommonHeaders(f func(http.ResponseWriter, *http.Request)) func(http.ResponseWriter, *http.Request) {
@@ -398,24 +631,6 @@ func (s *ProxyServer) setCommonHeaders(f func(http.ResponseWriter, *http.Request
 	}
 }
 
-func getRealUserIP(r *http.Request) string {
-	// Note this function assumes it is running behind a trusted proxy, so
-	// the headers can be trusted.
-	if ip := r.Header.Get("X-Real-IP"); ip != "" {
-		return ip
-	}
-
-	if ip := r.Header.Get("X-Forwarded-For"); ip != "" {
-		// Result could be a list "clientip, proxy1, proxy2", so only use first element.
-		if pos := strings.Index(ip, ","); pos >= 0 {
-			ip = strings.TrimSpace(ip[:pos])
-		}
-		return ip
-	}
-
-	return r.RemoteAddr
-}
-
 func (s *ProxyServer) welcomeHandler(w http.ResponseWriter, r *http.Request) {
 	w.Header().Set("Content-Type", "application/json; charset=utf-8")
 	w.WriteHeader(http.StatusOK)
@@ -423,21 +638,24 @@ func (s *ProxyServer) welcomeHandler(w http.ResponseWriter, r *http.Request) {
 }
 
 func (s *ProxyServer) proxyHandler(w http.ResponseWriter, r *http.Request) {
-	addr := getRealUserIP(r)
-	conn, err := s.upgrader.Upgrade(w, r, nil)
+	addr := signaling.GetRealUserIP(r, s.trustedProxies.Load())
+	header := http.Header{}
+	header.Set("Server", "nextcloud-spreed-signaling-proxy/"+s.version)
+	header.Set("X-Spreed-Signaling-Features", strings.Join(s.welcomeMsg.Features, ", "))
+	conn, err := s.upgrader.Upgrade(w, r, header)
 	if err != nil {
 		log.Printf("Could not upgrade request from %s: %s", addr, err)
 		return
 	}
 
-	client, err := NewProxyClient(s, conn, addr)
+	client, err := NewProxyClient(r.Context(), s, conn, addr)
 	if err != nil {
 		log.Printf("Could not create client for %s: %s", addr, err)
 		return
 	}
 
 	go client.WritePump()
-	go client.ReadPump()
+	client.ReadPump()
 }
 
 func (s *ProxyServer) clientClosed(client *signaling.Client) {
@@ -479,13 +697,7 @@ func (s *ProxyServer) onMcuDisconnected() {
 }
 
 func (s *ProxyServer) sendCurrentLoad(session *ProxySession) {
-	msg := &signaling.ProxyServerMessage{
-		Type: "event",
-		Event: &signaling.EventProxyServerMessage{
-			Type: "update-load",
-			Load: s.load.Load(),
-		},
-	}
+	msg := s.newLoadEvent(s.load.Load(), s.currentIncoming.Load(), s.currentOutgoing.Load())
 	session.sendMessage(msg)
 }
 
@@ -533,8 +745,7 @@ func (s *ProxyServer) processMessage(client *ProxyClient, data []byte) {
 
 		var session *ProxySession
 		if resumeId := message.Hello.ResumeId; resumeId != "" {
-			var data signaling.SessionIdData
-			if s.cookie.Decode("session", resumeId, &data) == nil {
+			if data, err := s.cookie.DecodePublic(resumeId); err == nil {
 				session = s.GetSession(data.Sid)
 			}
 
@@ -579,10 +790,7 @@ func (s *ProxyServer) processMessage(client *ProxyClient, data []byte) {
 			Hello: &signaling.HelloProxyServerMessage{
 				Version:   signaling.HelloVersionV1,
 				SessionId: session.PublicId(),
-				Server: &signaling.WelcomeServerMessage{
-					Version: s.version,
-					Country: s.country,
-				},
+				Server:    s.welcomeMsg,
 			},
 		}
 		client.SendMessage(response)
@@ -594,7 +802,7 @@ func (s *ProxyServer) processMessage(client *ProxyClient, data []byte) {
 		return
 	}
 
-	ctx := context.WithValue(context.Background(), ContextKeySession, session)
+	ctx := context.WithValue(session.Context(), ContextKeySession, session)
 	session.MarkUsed()
 
 	switch message.Type {
@@ -602,6 +810,8 @@ func (s *ProxyServer) processMessage(client *ProxyClient, data []byte) {
 		s.processCommand(ctx, client, session, &message)
 	case "payload":
 		s.processPayload(ctx, client, session, &message)
+	case "bye":
+		s.processBye(ctx, client, session, &message)
 	default:
 		session.sendMessage(message.NewErrorServerMessage(UnsupportedMessage))
 	}
@@ -613,6 +823,81 @@ func (i *emptyInitiator) Country() string {
 	return ""
 }
 
+type proxyRemotePublisher struct {
+	proxy     *ProxyServer
+	remoteUrl string
+
+	publisherId string
+}
+
+func (p *proxyRemotePublisher) PublisherId() string {
+	return p.publisherId
+}
+
+func (p *proxyRemotePublisher) StartPublishing(ctx context.Context, publisher signaling.McuRemotePublisherProperties) error {
+	conn, err := p.proxy.getRemoteConnection(p.remoteUrl)
+	if err != nil {
+		return err
+	}
+
+	if _, err := conn.RequestMessage(ctx, &signaling.ProxyClientMessage{
+		Type: "command",
+		Command: &signaling.CommandProxyClientMessage{
+			Type:     "publish-remote",
+			ClientId: p.publisherId,
+			Hostname: p.proxy.remoteHostname,
+			Port:     publisher.Port(),
+			RtcpPort: publisher.RtcpPort(),
+		},
+	}); err != nil {
+		return err
+	}
+
+	return nil
+}
+
+func (p *proxyRemotePublisher) StopPublishing(ctx context.Context, publisher signaling.McuRemotePublisherProperties) error {
+	conn, err := p.proxy.getRemoteConnection(p.remoteUrl)
+	if err != nil {
+		return err
+	}
+
+	if _, err := conn.RequestMessage(ctx, &signaling.ProxyClientMessage{
+		Type: "command",
+		Command: &signaling.CommandProxyClientMessage{
+			Type:     "unpublish-remote",
+			ClientId: p.publisherId,
+			Hostname: p.proxy.remoteHostname,
+			Port:     publisher.Port(),
+			RtcpPort: publisher.RtcpPort(),
+		},
+	}); err != nil {
+		return err
+	}
+
+	return nil
+}
+
+func (p *proxyRemotePublisher) GetStreams(ctx context.Context) ([]signaling.PublisherStream, error) {
+	conn, err := p.proxy.getRemoteConnection(p.remoteUrl)
+	if err != nil {
+		return nil, err
+	}
+
+	response, err := conn.RequestMessage(ctx, &signaling.ProxyClientMessage{
+		Type: "command",
+		Command: &signaling.CommandProxyClientMessage{
+			Type:     "get-publisher-streams",
+			ClientId: p.publisherId,
+		},
+	})
+	if err != nil {
+		return nil, err
+	}
+
+	return response.Command.Streams, nil
+}
+
 func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, session *ProxySession, message *signaling.ProxyClientMessage) {
 	cmd := message.Command
 
@@ -625,8 +910,18 @@ func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, s
 			return
 		}
 
+		ctx2, cancel := context.WithTimeout(ctx, s.mcuTimeout)
+		defer cancel()
+
 		id := uuid.New().String()
-		publisher, err := s.mcu.NewPublisher(ctx, session, id, cmd.Sid, cmd.StreamType, cmd.Bitrate, cmd.MediaTypes, &emptyInitiator{})
+		settings := cmd.PublisherSettings
+		if settings == nil {
+			settings = &signaling.NewPublisherSettings{
+				Bitrate:    cmd.Bitrate,    // nolint
+				MediaTypes: cmd.MediaTypes, // nolint
+			}
+		}
+		publisher, err := s.mcu.NewPublisher(ctx2, session, id, cmd.Sid, cmd.StreamType, *settings, &emptyInitiator{})
 		if err == context.DeadlineExceeded {
 			log.Printf("Timeout while creating %s publisher %s for %s", cmd.StreamType, id, session.PublicId())
 			session.sendMessage(message.NewErrorServerMessage(TimeoutCreatingPublisher))
@@ -655,18 +950,92 @@ func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, s
 	case "create-subscriber":
 		id := uuid.New().String()
 		publisherId := cmd.PublisherId
-		subscriber, err := s.mcu.NewSubscriber(ctx, session, publisherId, cmd.StreamType)
-		if err == context.DeadlineExceeded {
-			log.Printf("Timeout while creating %s subscriber on %s for %s", cmd.StreamType, publisherId, session.PublicId())
-			session.sendMessage(message.NewErrorServerMessage(TimeoutCreatingSubscriber))
-			return
-		} else if err != nil {
+		var subscriber signaling.McuSubscriber
+		var err error
+
+		handleCreateError := func(err error) {
+			if err == context.DeadlineExceeded {
+				log.Printf("Timeout while creating %s subscriber on %s for %s", cmd.StreamType, publisherId, session.PublicId())
+				session.sendMessage(message.NewErrorServerMessage(TimeoutCreatingSubscriber))
+				return
+			} else if errors.Is(err, signaling.ErrRemoteStreamsNotSupported) {
+				session.sendMessage(message.NewErrorServerMessage(RemoteSubscribersNotSupported))
+				return
+			}
+
 			log.Printf("Error while creating %s subscriber on %s for %s: %s", cmd.StreamType, publisherId, session.PublicId(), err)
 			session.sendMessage(message.NewWrappedErrorServerMessage(err))
-			return
 		}
 
-		log.Printf("Created %s subscriber %s as %s for %s", cmd.StreamType, subscriber.Id(), id, session.PublicId())
+		if cmd.RemoteUrl != "" {
+			if s.tokenId == "" || s.tokenKey == nil || s.remoteHostname == "" {
+				session.sendMessage(message.NewErrorServerMessage(RemoteSubscribersNotSupported))
+				return
+			}
+
+			remoteMcu, ok := s.mcu.(signaling.RemoteMcu)
+			if !ok {
+				session.sendMessage(message.NewErrorServerMessage(RemoteSubscribersNotSupported))
+				return
+			}
+
+			claims, _, err := s.parseToken(cmd.RemoteToken)
+			if err != nil {
+				if e, ok := err.(*signaling.Error); ok {
+					client.SendMessage(message.NewErrorServerMessage(e))
+				} else {
+					client.SendMessage(message.NewWrappedErrorServerMessage(err))
+				}
+				return
+			}
+
+			if claims.Subject != publisherId {
+				session.sendMessage(message.NewErrorServerMessage(TokenAuthFailed))
+				return
+			}
+
+			subCtx, cancel := context.WithTimeout(ctx, remotePublisherTimeout)
+			defer cancel()
+
+			log.Printf("Creating remote subscriber for %s on %s", publisherId, cmd.RemoteUrl)
+
+			controller := &proxyRemotePublisher{
+				proxy:       s,
+				remoteUrl:   cmd.RemoteUrl,
+				publisherId: publisherId,
+			}
+
+			var publisher signaling.McuRemotePublisher
+			publisher, err = remoteMcu.NewRemotePublisher(subCtx, session, controller, cmd.StreamType)
+			if err != nil {
+				handleCreateError(err)
+				return
+			}
+
+			defer func() {
+				go publisher.Close(context.Background())
+			}()
+
+			subscriber, err = remoteMcu.NewRemoteSubscriber(subCtx, session, publisher)
+			if err != nil {
+				handleCreateError(err)
+				return
+			}
+
+			log.Printf("Created remote %s subscriber %s as %s for %s on %s", cmd.StreamType, subscriber.Id(), id, session.PublicId(), cmd.RemoteUrl)
+		} else {
+			ctx2, cancel := context.WithTimeout(ctx, s.mcuTimeout)
+			defer cancel()
+
+			subscriber, err = s.mcu.NewSubscriber(ctx2, session, publisherId, cmd.StreamType, &emptyInitiator{})
+			if err != nil {
+				handleCreateError(err)
+				return
+			}
+
+			log.Printf("Created %s subscriber %s as %s for %s", cmd.StreamType, subscriber.Id(), id, session.PublicId())
+		}
+
 		session.StoreSubscriber(ctx, id, subscriber)
 		s.StoreClient(id, subscriber)
 
@@ -751,6 +1120,119 @@ func (s *ProxyServer) processCommand(ctx context.Context, client *ProxyClient, s
 			},
 		}
 		session.sendMessage(response)
+	case "publish-remote":
+		client := s.GetClient(cmd.ClientId)
+		if client == nil {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		publisher, ok := client.(signaling.McuPublisher)
+		if !ok {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		ctx2, cancel := context.WithTimeout(ctx, s.mcuTimeout)
+		defer cancel()
+
+		if err := publisher.PublishRemote(ctx2, session.PublicId(), cmd.Hostname, cmd.Port, cmd.RtcpPort); err != nil {
+			var je *janus.ErrorMsg
+			if !errors.As(err, &je) || je.Err.Code != signaling.JANUS_VIDEOROOM_ERROR_ID_EXISTS {
+				log.Printf("Error publishing %s %s to remote %s (port=%d, rtcpPort=%d): %s", publisher.StreamType(), cmd.ClientId, cmd.Hostname, cmd.Port, cmd.RtcpPort, err)
+				session.sendMessage(message.NewWrappedErrorServerMessage(err))
+				return
+			}
+
+			ctx2, cancel = context.WithTimeout(ctx, s.mcuTimeout)
+			defer cancel()
+
+			if err := publisher.UnpublishRemote(ctx2, session.PublicId(), cmd.Hostname, cmd.Port, cmd.RtcpPort); err != nil {
+				log.Printf("Error unpublishing old %s %s to remote %s (port=%d, rtcpPort=%d): %s", publisher.StreamType(), cmd.ClientId, cmd.Hostname, cmd.Port, cmd.RtcpPort, err)
+				session.sendMessage(message.NewWrappedErrorServerMessage(err))
+				return
+			}
+
+			ctx2, cancel = context.WithTimeout(ctx, s.mcuTimeout)
+			defer cancel()
+
+			if err := publisher.PublishRemote(ctx2, session.PublicId(), cmd.Hostname, cmd.Port, cmd.RtcpPort); err != nil {
+				log.Printf("Error publishing %s %s to remote %s (port=%d, rtcpPort=%d): %s", publisher.StreamType(), cmd.ClientId, cmd.Hostname, cmd.Port, cmd.RtcpPort, err)
+				session.sendMessage(message.NewWrappedErrorServerMessage(err))
+				return
+			}
+		}
+
+		session.AddRemotePublisher(publisher, cmd.Hostname, cmd.Port, cmd.RtcpPort)
+		response := &signaling.ProxyServerMessage{
+			Id:   message.Id,
+			Type: "command",
+			Command: &signaling.CommandProxyServerMessage{
+				Id: cmd.ClientId,
+			},
+		}
+		session.sendMessage(response)
+	case "unpublish-remote":
+		client := s.GetClient(cmd.ClientId)
+		if client == nil {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		publisher, ok := client.(signaling.McuPublisher)
+		if !ok {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		ctx2, cancel := context.WithTimeout(ctx, s.mcuTimeout)
+		defer cancel()
+
+		if err := publisher.UnpublishRemote(ctx2, session.PublicId(), cmd.Hostname, cmd.Port, cmd.RtcpPort); err != nil {
+			log.Printf("Error unpublishing %s %s from remote %s: %s", publisher.StreamType(), cmd.ClientId, cmd.Hostname, err)
+			session.sendMessage(message.NewWrappedErrorServerMessage(err))
+			return
+		}
+
+		session.RemoveRemotePublisher(publisher, cmd.Hostname, cmd.Port, cmd.RtcpPort)
+
+		response := &signaling.ProxyServerMessage{
+			Id:   message.Id,
+			Type: "command",
+			Command: &signaling.CommandProxyServerMessage{
+				Id: cmd.ClientId,
+			},
+		}
+		session.sendMessage(response)
+	case "get-publisher-streams":
+		client := s.GetClient(cmd.ClientId)
+		if client == nil {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		publisher, ok := client.(signaling.McuPublisher)
+		if !ok {
+			session.sendMessage(message.NewErrorServerMessage(UnknownClient))
+			return
+		}
+
+		streams, err := publisher.GetStreams(ctx)
+		if err != nil {
+			log.Printf("Could not get streams of publisher %s: %s", publisher.Id(), err)
+			session.sendMessage(message.NewWrappedErrorServerMessage(err))
+			return
+		}
+
+		response := &signaling.ProxyServerMessage{
+			Id:   message.Id,
+			Type: "command",
+			Command: &signaling.CommandProxyServerMessage{
+				Id:      cmd.ClientId,
+				Streams: streams,
+			},
+		}
+		session.sendMessage(response)
 	default:
 		log.Printf("Unsupported command %+v", message.Command)
 		session.sendMessage(message.NewErrorServerMessage(UnsupportedCommand))
@@ -777,9 +1259,10 @@ func (s *ProxyServer) processPayload(ctx context.Context, client *ProxyClient, s
 		fallthrough
 	case "candidate":
 		mcuData = &signaling.MessageClientMessageData{
-			Type:    payload.Type,
-			Sid:     payload.Sid,
-			Payload: payload.Payload,
+			RoomType: string(mcuClient.StreamType()),
+			Type:     payload.Type,
+			Sid:      payload.Sid,
+			Payload:  payload.Payload,
 		}
 	case "endOfCandidates":
 		// Ignore but confirm, not passed along to Janus anyway.
@@ -796,15 +1279,25 @@ func (s *ProxyServer) processPayload(ctx context.Context, client *ProxyClient, s
 		fallthrough
 	case "sendoffer":
 		mcuData = &signaling.MessageClientMessageData{
-			Type: payload.Type,
-			Sid:  payload.Sid,
+			RoomType: string(mcuClient.StreamType()),
+			Type:     payload.Type,
+			Sid:      payload.Sid,
 		}
 	default:
 		session.sendMessage(message.NewErrorServerMessage(UnsupportedPayload))
 		return
 	}
 
-	mcuClient.SendMessage(ctx, nil, mcuData, func(err error, response map[string]interface{}) {
+	if err := mcuData.CheckValid(); err != nil {
+		log.Printf("Received invalid payload %+v for %s client %s: %s", mcuData, mcuClient.StreamType(), payload.ClientId, err)
+		session.sendMessage(message.NewErrorServerMessage(UnsupportedPayload))
+		return
+	}
+
+	ctx2, cancel := context.WithTimeout(ctx, s.mcuTimeout)
+	defer cancel()
+
+	mcuClient.SendMessage(ctx2, nil, mcuData, func(err error, response map[string]interface{}) {
 		var responseMsg *signaling.ProxyServerMessage
 		if err != nil {
 			log.Printf("Error sending %+v to %s client %s: %s", mcuData, mcuClient.StreamType(), payload.ClientId, err)
@@ -825,13 +1318,14 @@ func (s *ProxyServer) processPayload(ctx context.Context, client *ProxyClient, s
 	})
 }
 
-func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*ProxySession, error) {
-	if proxyDebugMessages {
-		log.Printf("Hello: %+v", hello)
-	}
+func (s *ProxyServer) processBye(ctx context.Context, client *ProxyClient, session *ProxySession, message *signaling.ProxyClientMessage) {
+	log.Printf("Closing session %s", session.PublicId())
+	s.DeleteSession(session.Sid())
+}
 
+func (s *ProxyServer) parseToken(tokenValue string) (*signaling.TokenClaims, string, error) {
 	reason := "auth-failed"
-	token, err := jwt.ParseWithClaims(hello.Token, &signaling.TokenClaims{}, func(token *jwt.Token) (interface{}, error) {
+	token, err := jwt.ParseWithClaims(tokenValue, &signaling.TokenClaims{}, func(token *jwt.Token) (interface{}, error) {
 		// Don't forget to validate the alg is what you expect:
 		if _, ok := token.Method.(*jwt.SigningMethodRSA); !ok {
 			log.Printf("Unexpected signing method: %v", token.Header["alg"])
@@ -860,28 +1354,44 @@ func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*Pro
 		}
 
 		return tokenKey.key, nil
-	})
-	if err, ok := err.(*jwt.ValidationError); ok {
-		if err.Errors&jwt.ValidationErrorIssuedAt == jwt.ValidationErrorIssuedAt {
-			statsTokenErrorsTotal.WithLabelValues("not-valid-yet").Inc()
-			return nil, TokenNotValidYet
-		}
-	}
+	}, jwt.WithValidMethods([]string{
+		jwt.SigningMethodRS256.Alg(),
+		jwt.SigningMethodRS384.Alg(),
+		jwt.SigningMethodRS512.Alg(),
+	}), jwt.WithIssuedAt(), jwt.WithLeeway(tokenLeeway))
 	if err != nil {
-		statsTokenErrorsTotal.WithLabelValues(reason).Inc()
-		return nil, TokenAuthFailed
+		if errors.Is(err, jwt.ErrTokenNotValidYet) || errors.Is(err, jwt.ErrTokenUsedBeforeIssued) {
+			return nil, "not-valid-yet", TokenNotValidYet
+		} else if errors.Is(err, jwt.ErrTokenExpired) {
+			return nil, "expired", TokenExpired
+		}
+
+		return nil, reason, TokenAuthFailed
 	}
 
 	claims, ok := token.Claims.(*signaling.TokenClaims)
 	if !ok || !token.Valid {
-		statsTokenErrorsTotal.WithLabelValues("auth-failed").Inc()
-		return nil, TokenAuthFailed
+		return nil, "auth-failed", TokenAuthFailed
 	}
 
-	minIssuedAt := time.Now().Add(-maxTokenAge)
-	if issuedAt := claims.IssuedAt; issuedAt != nil && issuedAt.Before(minIssuedAt) {
-		statsTokenErrorsTotal.WithLabelValues("expired").Inc()
-		return nil, TokenExpired
+	now := time.Now()
+	minIssuedAt := now.Add(-(maxTokenAge + tokenLeeway))
+	if issuedAt := claims.IssuedAt; issuedAt == nil || issuedAt.Before(minIssuedAt) {
+		return nil, "expired", TokenExpired
+	}
+
+	return claims, "", nil
+}
+
+func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*ProxySession, error) {
+	if proxyDebugMessages {
+		log.Printf("Hello: %+v", hello)
+	}
+
+	claims, reason, err := s.parseToken(hello.Token)
+	if err != nil {
+		statsTokenErrorsTotal.WithLabelValues(reason).Inc()
+		return nil, err
 	}
 
 	sid := s.sid.Add(1)
@@ -891,10 +1401,10 @@ func (s *ProxyServer) NewSession(hello *signaling.HelloProxyClientMessage) (*Pro
 
 	sessionIdData := &signaling.SessionIdData{
 		Sid:     sid,
-		Created: time.Now(),
+		Created: timestamppb.Now(),
 	}
 
-	encoded, err := s.cookie.Encode("session", sessionIdData)
+	encoded, err := s.cookie.EncodePublic(sessionIdData)
 	if err != nil {
 		return nil, err
 	}
@@ -943,6 +1453,8 @@ func (s *ProxyServer) DeleteSession(id uint64) {
 func (s *ProxyServer) deleteSessionLocked(id uint64) {
 	if session, found := s.sessions[id]; found {
 		delete(s.sessions, id)
+		s.sessionsLock.Unlock()
+		defer s.sessionsLock.Lock()
 		session.Close()
 		statsSessionsCurrent.Dec()
 	}
@@ -977,15 +1489,21 @@ func (s *ProxyServer) HasClients() bool {
 	return len(s.clients) > 0
 }
 
-func (s *ProxyServer) GetClientsLoad() int64 {
+func (s *ProxyServer) GetClientsLoad() (load int64, incoming int64, outgoing int64) {
 	s.clientsLock.RLock()
 	defer s.clientsLock.RUnlock()
 
-	var load int64
 	for _, c := range s.clients {
-		load += int64(c.MaxBitrate())
+		bitrate := int64(c.MaxBitrate())
+		load += bitrate
+		if _, ok := c.(signaling.McuPublisher); ok {
+			incoming += bitrate
+		} else if _, ok := c.(signaling.McuSubscriber); ok {
+			outgoing += bitrate
+		}
 	}
-	return load / 1024
+	load = load / 1024
+	return
 }
 
 func (s *ProxyServer) GetClient(id string) signaling.McuClient {
@@ -994,6 +1512,22 @@ func (s *ProxyServer) GetClient(id string) signaling.McuClient {
 	return s.clients[id]
 }
 
+func (s *ProxyServer) GetPublisher(publisherId string) signaling.McuPublisher {
+	s.clientsLock.RLock()
+	defer s.clientsLock.RUnlock()
+	for _, c := range s.clients {
+		pub, ok := c.(signaling.McuPublisher)
+		if !ok {
+			continue
+		}
+
+		if pub.Id() == publisherId {
+			return pub
+		}
+	}
+	return nil
+}
+
 func (s *ProxyServer) GetClientId(client signaling.McuClient) string {
 	s.clientsLock.RLock()
 	defer s.clientsLock.RUnlock()
@@ -1010,19 +1544,14 @@ func (s *ProxyServer) getStats() map[string]interface{} {
 }
 
 func (s *ProxyServer) allowStatsAccess(r *http.Request) bool {
-	addr := getRealUserIP(r)
-	if strings.Contains(addr, ":") {
-		if host, _, err := net.SplitHostPort(addr); err == nil {
-			addr = host
-		}
-	}
-
+	addr := signaling.GetRealUserIP(r, s.trustedProxies.Load())
 	ip := net.ParseIP(addr)
-	if ip == nil {
+	if len(ip) == 0 {
 		return false
 	}
 
-	return s.statsAllowedIps.Allowed(ip)
+	allowed := s.statsAllowedIps.Load()
+	return allowed != nil && allowed.Allowed(ip)
 }
 
 func (s *ProxyServer) validateStatsRequest(f func(http.ResponseWriter, *http.Request)) func(http.ResponseWriter, *http.Request) {
@@ -1055,3 +1584,30 @@ func (s *ProxyServer) metricsHandler(w http.ResponseWriter, r *http.Request) {
 	// Expose prometheus metrics at "/metrics".
 	promhttp.Handler().ServeHTTP(w, r)
 }
+
+func (s *ProxyServer) getRemoteConnection(url string) (*RemoteConnection, error) {
+	s.remoteConnectionsLock.Lock()
+	defer s.remoteConnectionsLock.Unlock()
+
+	conn, found := s.remoteConnections[url]
+	if found {
+		return conn, nil
+	}
+
+	conn, err := NewRemoteConnection(url, s.tokenId, s.tokenKey, s.remoteTlsConfig)
+	if err != nil {
+		return nil, err
+	}
+
+	s.remoteConnections[url] = conn
+	return conn, nil
+}
+
+func (s *ProxyServer) PublisherDeleted(publisher signaling.McuPublisher) {
+	s.sessionsLock.RLock()
+	defer s.sessionsLock.RUnlock()
+
+	for _, session := range s.sessions {
+		session.OnPublisherDeleted(publisher)
+	}
+}
