@@ -23,6 +23,7 @@ package main
 
 import (
 	"context"
+	"fmt"
 	"log"
 	"sync"
 	"sync/atomic"
@@ -36,11 +37,19 @@ const (
 	sessionExpirationTime = time.Minute
 )
 
+type remotePublisherData struct {
+	hostname string
+	port     int
+	rtcpPort int
+}
+
 type ProxySession struct {
-	proxy    *ProxyServer
-	id       string
-	sid      uint64
-	lastUsed atomic.Int64
+	proxy     *ProxyServer
+	id        string
+	sid       uint64
+	lastUsed  atomic.Int64
+	ctx       context.Context
+	closeFunc context.CancelFunc
 
 	clientLock      sync.Mutex
 	client          *ProxyClient
@@ -53,13 +62,19 @@ type ProxySession struct {
 	subscribersLock sync.Mutex
 	subscribers     map[string]signaling.McuSubscriber
 	subscriberIds   map[signaling.McuSubscriber]string
+
+	remotePublishersLock sync.Mutex
+	remotePublishers     map[signaling.McuPublisher]map[string]*remotePublisherData
 }
 
 func NewProxySession(proxy *ProxyServer, sid uint64, id string) *ProxySession {
+	ctx, closeFunc := context.WithCancel(context.Background())
 	result := &ProxySession{
-		proxy: proxy,
-		id:    id,
-		sid:   sid,
+		proxy:     proxy,
+		id:        id,
+		sid:       sid,
+		ctx:       ctx,
+		closeFunc: closeFunc,
 
 		publishers:   make(map[string]signaling.McuPublisher),
 		publisherIds: make(map[signaling.McuPublisher]string),
@@ -71,6 +86,10 @@ func NewProxySession(proxy *ProxyServer, sid uint64, id string) *ProxySession {
 	return result
 }
 
+func (s *ProxySession) Context() context.Context {
+	return s.ctx
+}
+
 func (s *ProxySession) PublicId() string {
 	return s.id
 }
@@ -95,8 +114,25 @@ func (s *ProxySession) MarkUsed() {
 }
 
 func (s *ProxySession) Close() {
+	prev := s.SetClient(nil)
+	if prev != nil {
+		reason := "session_closed"
+		if s.IsExpired() {
+			reason = "session_expired"
+		}
+		prev.SendMessage(&signaling.ProxyServerMessage{
+			Type: "bye",
+			Bye: &signaling.ByeProxyServerMessage{
+				Reason: reason,
+			},
+		})
+	}
+
+	s.closeFunc()
 	s.clearPublishers()
 	s.clearSubscribers()
+	s.clearRemotePublishers()
+	s.proxy.DeleteSession(s.Sid())
 }
 
 func (s *ProxySession) SetClient(client *ProxyClient) *ProxyClient {
@@ -262,6 +298,8 @@ func (s *ProxySession) DeletePublisher(publisher signaling.McuPublisher) string
 
 	delete(s.publishers, id)
 	delete(s.publisherIds, publisher)
+	delete(s.remotePublishers, publisher)
+	go s.proxy.PublisherDeleted(publisher)
 	return id
 }
 
@@ -299,8 +337,25 @@ func (s *ProxySession) clearPublishers() {
 			publisher.Close(context.Background())
 		}
 	}(s.publishers)
+	// Can't use clear(...) here as the map is processed by the goroutine above.
 	s.publishers = make(map[string]signaling.McuPublisher)
-	s.publisherIds = make(map[signaling.McuPublisher]string)
+	clear(s.publisherIds)
+}
+
+func (s *ProxySession) clearRemotePublishers() {
+	s.remotePublishersLock.Lock()
+	defer s.remotePublishersLock.Unlock()
+
+	go func(remotePublishers map[signaling.McuPublisher]map[string]*remotePublisherData) {
+		for publisher, entries := range remotePublishers {
+			for _, data := range entries {
+				if err := publisher.UnpublishRemote(context.Background(), s.PublicId(), data.hostname, data.port, data.rtcpPort); err != nil {
+					log.Printf("Error unpublishing %s %s from remote %s: %s", publisher.StreamType(), publisher.Id(), data.hostname, err)
+				}
+			}
+		}
+	}(s.remotePublishers)
+	s.remotePublishers = nil
 }
 
 func (s *ProxySession) clearSubscribers() {
@@ -315,11 +370,66 @@ func (s *ProxySession) clearSubscribers() {
 			subscriber.Close(context.Background())
 		}
 	}(s.subscribers)
+	// Can't use clear(...) here as the map is processed by the goroutine above.
 	s.subscribers = make(map[string]signaling.McuSubscriber)
-	s.subscriberIds = make(map[signaling.McuSubscriber]string)
+	clear(s.subscriberIds)
 }
 
 func (s *ProxySession) NotifyDisconnected() {
 	s.clearPublishers()
 	s.clearSubscribers()
+	s.clearRemotePublishers()
+}
+
+func (s *ProxySession) AddRemotePublisher(publisher signaling.McuPublisher, hostname string, port int, rtcpPort int) bool {
+	s.remotePublishersLock.Lock()
+	defer s.remotePublishersLock.Unlock()
+
+	remote, found := s.remotePublishers[publisher]
+	if !found {
+		remote = make(map[string]*remotePublisherData)
+		if s.remotePublishers == nil {
+			s.remotePublishers = make(map[signaling.McuPublisher]map[string]*remotePublisherData)
+		}
+		s.remotePublishers[publisher] = remote
+	}
+
+	key := fmt.Sprintf("%s:%d%d", hostname, port, rtcpPort)
+	if _, found := remote[key]; found {
+		return false
+	}
+
+	data := &remotePublisherData{
+		hostname: hostname,
+		port:     port,
+		rtcpPort: rtcpPort,
+	}
+	remote[key] = data
+	return true
+}
+
+func (s *ProxySession) RemoveRemotePublisher(publisher signaling.McuPublisher, hostname string, port int, rtcpPort int) {
+	s.remotePublishersLock.Lock()
+	defer s.remotePublishersLock.Unlock()
+
+	remote, found := s.remotePublishers[publisher]
+	if !found {
+		return
+	}
+
+	key := fmt.Sprintf("%s:%d%d", hostname, port, rtcpPort)
+	delete(remote, key)
+	if len(remote) == 0 {
+		delete(s.remotePublishers, publisher)
+		if len(s.remotePublishers) == 0 {
+			s.remotePublishers = nil
+		}
+	}
+}
+
+func (s *ProxySession) OnPublisherDeleted(publisher signaling.McuPublisher) {
+	s.remotePublishersLock.Lock()
+	defer s.remotePublishersLock.Unlock()
+
+	delete(s.remotePublishers, publisher)
 }
