@@ -29,6 +29,8 @@ import (
 	"strings"
 	"sync"
 	"testing"
+
+	"github.com/stretchr/testify/assert"
 )
 
 var (
@@ -61,9 +63,7 @@ func newMcuProxyForConfig(t *testing.T) *mcuProxyForConfig {
 		t: t,
 	}
 	t.Cleanup(func() {
-		if len(proxy.expected) > 0 {
-			t.Errorf("expected events %+v were not triggered", proxy.expected)
-		}
+		assert.Empty(t, proxy.expected)
 	})
 	return proxy
 }
@@ -99,7 +99,7 @@ func (p *mcuProxyForConfig) WaitForEvents(ctx context.Context) {
 	defer p.mu.Lock()
 	select {
 	case <-ctx.Done():
-		p.t.Error(ctx.Err())
+		assert.NoError(p.t, ctx.Err())
 	case <-waiter:
 	}
 }
@@ -125,7 +125,7 @@ func (p *mcuProxyForConfig) checkEvent(event *proxyConfigEvent) {
 	defer p.mu.Unlock()
 
 	if len(p.expected) == 0 {
-		p.t.Errorf("no event expected, got %+v from %s:%d", event, caller.File, caller.Line)
+		assert.Fail(p.t, "no event expected, got %+v from %s:%d", event, caller.File, caller.Line)
 		return
 	}
 
@@ -145,7 +145,7 @@ func (p *mcuProxyForConfig) checkEvent(event *proxyConfigEvent) {
 	expected := p.expected[0]
 	p.expected = p.expected[1:]
 	if !reflect.DeepEqual(expected, *event) {
-		p.t.Errorf("expected %+v, got %+v from %s:%d", expected, event, caller.File, caller.Line)
+		assert.Fail(p.t, "expected %+v, got %+v from %s:%d", expected, event, caller.File, caller.Line)
 	}
 }
 
