@@ -65,13 +65,13 @@ type Room struct {
 	events  AsyncEvents
 	backend *Backend
 
-	properties *json.RawMessage
+	properties json.RawMessage
 
 	closer   *Closer
 	mu       *sync.RWMutex
 	sessions map[string]Session
 
-	internalSessions map[Session]bool
+	internalSessions map[*ClientSession]bool
 	virtualSessions  map[*VirtualSession]bool
 	inCallSessions   map[Session]bool
 	roomSessionData  map[string]*RoomSessionData
@@ -95,7 +95,7 @@ func getRoomIdForBackend(id string, backend *Backend) string {
 	return backend.Id() + "|" + id
 }
 
-func NewRoom(roomId string, properties *json.RawMessage, hub *Hub, events AsyncEvents, backend *Backend) (*Room, error) {
+func NewRoom(roomId string, properties json.RawMessage, hub *Hub, events AsyncEvents, backend *Backend) (*Room, error) {
 	room := &Room{
 		id:      roomId,
 		hub:     hub,
@@ -108,7 +108,7 @@ func NewRoom(roomId string, properties *json.RawMessage, hub *Hub, events AsyncE
 		mu:       &sync.RWMutex{},
 		sessions: make(map[string]Session),
 
-		internalSessions: make(map[Session]bool),
+		internalSessions: make(map[*ClientSession]bool),
 		virtualSessions:  make(map[*VirtualSession]bool),
 		inCallSessions:   make(map[Session]bool),
 		roomSessionData:  make(map[string]*RoomSessionData),
@@ -136,7 +136,7 @@ func (r *Room) Id() string {
 	return r.id
 }
 
-func (r *Room) Properties() *json.RawMessage {
+func (r *Room) Properties() json.RawMessage {
 	r.mu.RLock()
 	defer r.mu.RUnlock()
 	return r.properties
@@ -270,12 +270,12 @@ func (r *Room) processBackendRoomRequestAsyncRoom(message *AsyncRoomMessage) {
 	}
 }
 
-func (r *Room) AddSession(session Session, sessionData *json.RawMessage) {
+func (r *Room) AddSession(session Session, sessionData json.RawMessage) {
 	var roomSessionData *RoomSessionData
-	if sessionData != nil && len(*sessionData) > 0 {
+	if len(sessionData) > 0 {
 		roomSessionData = &RoomSessionData{}
-		if err := json.Unmarshal(*sessionData, roomSessionData); err != nil {
-			log.Printf("Error decoding room session data \"%s\": %s", string(*sessionData), err)
+		if err := json.Unmarshal(sessionData, roomSessionData); err != nil {
+			log.Printf("Error decoding room session data \"%s\": %s", string(sessionData), err)
 			roomSessionData = nil
 		}
 	}
@@ -290,13 +290,19 @@ func (r *Room) AddSession(session Session, sessionData *json.RawMessage) {
 	var publishUsersChanged bool
 	switch session.ClientType() {
 	case HelloClientTypeInternal:
-		r.internalSessions[session] = true
+		clientSession, ok := session.(*ClientSession)
+		if !ok {
+			delete(r.sessions, sid)
+			r.mu.Unlock()
+			panic(fmt.Sprintf("Expected a client session, got %v (%T)", session, session))
+		}
+		r.internalSessions[clientSession] = true
 	case HelloClientTypeVirtual:
 		virtualSession, ok := session.(*VirtualSession)
 		if !ok {
 			delete(r.sessions, sid)
 			r.mu.Unlock()
-			panic(fmt.Sprintf("Expected a virtual session, got %v", session))
+			panic(fmt.Sprintf("Expected a virtual session, got %v (%T)", session, session))
 		}
 		r.virtualSessions[virtualSession] = true
 		publishUsersChanged = true
@@ -366,7 +372,9 @@ func (r *Room) notifySessionJoined(sessionId string) {
 			User:      s.UserData(),
 		}
 		if s, ok := s.(*ClientSession); ok {
+			entry.Features = s.GetFeatures()
 			entry.RoomSessionId = s.RoomSessionId()
+			entry.Federated = s.ClientType() == HelloClientTypeFederation
 		}
 		events = append(events, entry)
 	}
@@ -446,11 +454,21 @@ func (r *Room) RemoveSession(session Session) bool {
 	sid := session.PublicId()
 	r.statsRoomSessionsCurrent.With(prometheus.Labels{"clienttype": session.ClientType()}).Dec()
 	delete(r.sessions, sid)
-	delete(r.internalSessions, session)
 	if virtualSession, ok := session.(*VirtualSession); ok {
 		delete(r.virtualSessions, virtualSession)
+		// Handle case where virtual session was also sent by Nextcloud.
+		users := make([]map[string]interface{}, 0, len(r.users))
+		for _, u := range r.users {
+			if u["sessionId"] != sid {
+				users = append(users, u)
+			}
+		}
+		if len(users) != len(r.users) {
+			r.users = users
+		}
 	}
 	if clientSession, ok := session.(*ClientSession); ok {
+		delete(r.internalSessions, clientSession)
 		r.transientData.RemoveListener(clientSession)
 	}
 	delete(r.inCallSessions, session)
@@ -480,11 +498,11 @@ func (r *Room) publish(message *ServerMessage) error {
 	})
 }
 
-func (r *Room) UpdateProperties(properties *json.RawMessage) {
+func (r *Room) UpdateProperties(properties json.RawMessage) {
 	r.mu.Lock()
 	defer r.mu.Unlock()
-	if (r.properties == nil && properties == nil) ||
-		(r.properties != nil && properties != nil && bytes.Equal(*r.properties, *properties)) {
+	if (len(r.properties) == 0 && len(properties) == 0) ||
+		(len(r.properties) > 0 && len(properties) > 0 && bytes.Equal(r.properties, properties)) {
 		// Don't notify if properties didn't change.
 		return
 	}
@@ -534,7 +552,9 @@ func (r *Room) PublishSessionJoined(session Session, sessionData *RoomSessionDat
 		},
 	}
 	if session, ok := session.(*ClientSession); ok {
+		message.Event.Join[0].Features = session.GetFeatures()
 		message.Event.Join[0].RoomSessionId = session.RoomSessionId()
+		message.Event.Join[0].Federated = session.ClientType() == HelloClientTypeFederation
 	}
 	if err := r.publish(message); err != nil {
 		log.Printf("Could not publish session joined message in room %s: %s", r.Id(), err)
@@ -566,10 +586,66 @@ func (r *Room) PublishSessionLeft(session Session) {
 	}
 }
 
+func (r *Room) getClusteredInternalSessionsRLocked() (internal map[string]*InternalSessionData, virtual map[string]*VirtualSessionData) {
+	if r.hub.rpcClients == nil {
+		return nil, nil
+	}
+
+	r.mu.RUnlock()
+	defer r.mu.RLock()
+	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
+	defer cancel()
+
+	var mu sync.Mutex
+	var wg sync.WaitGroup
+	for _, client := range r.hub.rpcClients.GetClients() {
+		wg.Add(1)
+		go func(c *GrpcClient) {
+			defer wg.Done()
+
+			clientInternal, clientVirtual, err := c.GetInternalSessions(ctx, r.Id(), r.Backend())
+			if err != nil {
+				log.Printf("Received error while getting internal sessions for %s@%s from %s: %s", r.Id(), r.Backend().Id(), c.Target(), err)
+				return
+			}
+
+			mu.Lock()
+			defer mu.Unlock()
+			if internal == nil {
+				internal = make(map[string]*InternalSessionData, len(clientInternal))
+			}
+			for sid, s := range clientInternal {
+				internal[sid] = s
+			}
+			if virtual == nil {
+				virtual = make(map[string]*VirtualSessionData, len(clientVirtual))
+			}
+			for sid, s := range clientVirtual {
+				virtual[sid] = s
+			}
+		}(client)
+	}
+	wg.Wait()
+
+	return
+}
+
 func (r *Room) addInternalSessions(users []map[string]interface{}) []map[string]interface{} {
 	now := time.Now().Unix()
-	r.mu.Lock()
-	defer r.mu.Unlock()
+	r.mu.RLock()
+	defer r.mu.RUnlock()
+	if len(users) == 0 && len(r.internalSessions) == 0 && len(r.virtualSessions) == 0 {
+		return users
+	}
+
+	clusteredInternalSessions, clusteredVirtualSessions := r.getClusteredInternalSessionsRLocked()
+
+	// Local sessions might have changed while waiting for clustered information.
+	if len(users) == 0 && len(r.internalSessions) == 0 && len(r.virtualSessions) == 0 {
+		return users
+	}
+
+	skipSession := make(map[string]bool)
 	for _, user := range users {
 		sessionid, found := user["sessionId"]
 		if !found || sessionid == "" {
@@ -579,21 +655,69 @@ func (r *Room) addInternalSessions(users []map[string]interface{}) []map[string]
 		if userid, found := user["userId"]; !found || userid == "" {
 			if roomSessionData, found := r.roomSessionData[sessionid.(string)]; found {
 				user["userId"] = roomSessionData.UserId
+			} else if sid, ok := sessionid.(string); ok {
+				if entry, found := clusteredVirtualSessions[sid]; found {
+					user["virtual"] = true
+					user["inCall"] = entry.GetInCall()
+					skipSession[sid] = true
+				} else {
+					for session := range r.virtualSessions {
+						if session.PublicId() == sid {
+							user["virtual"] = true
+							user["inCall"] = session.GetInCall()
+							skipSession[sid] = true
+							break
+						}
+					}
+				}
 			}
 		}
 	}
 	for session := range r.internalSessions {
-		users = append(users, map[string]interface{}{
-			"inCall":    session.(*ClientSession).GetInCall(),
+		u := map[string]interface{}{
+			"inCall":    session.GetInCall(),
 			"sessionId": session.PublicId(),
 			"lastPing":  now,
 			"internal":  true,
-		})
+		}
+		if f := session.GetFeatures(); len(f) > 0 {
+			u["features"] = f
+		}
+		users = append(users, u)
+	}
+	for _, session := range clusteredInternalSessions {
+		u := map[string]interface{}{
+			"inCall":    session.GetInCall(),
+			"sessionId": session.GetSessionId(),
+			"lastPing":  now,
+			"internal":  true,
+		}
+		if f := session.GetFeatures(); len(f) > 0 {
+			u["features"] = f
+		}
+		users = append(users, u)
 	}
 	for session := range r.virtualSessions {
+		sid := session.PublicId()
+		if skipSession[sid] {
+			continue
+		}
+		skipSession[sid] = true
 		users = append(users, map[string]interface{}{
 			"inCall":    session.GetInCall(),
-			"sessionId": session.PublicId(),
+			"sessionId": sid,
+			"lastPing":  now,
+			"virtual":   true,
+		})
+	}
+	for sid, session := range clusteredVirtualSessions {
+		if skipSession[sid] {
+			continue
+		}
+
+		users = append(users, map[string]interface{}{
+			"inCall":    session.GetInCall(),
+			"sessionId": sid,
 			"lastPing":  now,
 			"virtual":   true,
 		})
@@ -709,7 +833,8 @@ func (r *Room) PublishUsersInCallChangedAll(inCall int) {
 				continue
 			}
 
-			if session.ClientType() == HelloClientTypeInternal {
+			if session.ClientType() == HelloClientTypeInternal ||
+				session.ClientType() == HelloClientTypeFederation {
 				continue
 			}
 
@@ -769,7 +894,7 @@ func (r *Room) PublishUsersInCallChangedAll(inCall int) {
 			Type:   "update",
 			Update: &RoomEventServerMessage{
 				RoomId: r.id,
-				InCall: &inCallMsg,
+				InCall: inCallMsg,
 				All:    true,
 			},
 		},
@@ -953,23 +1078,25 @@ func (r *Room) publishActiveSessions() (int, *sync.WaitGroup) {
 	if len(urls) == 0 {
 		return 0, &wg
 	}
+	var count int
 	for u, e := range entries {
 		wg.Add(1)
+		count += len(e)
 		go func(url *url.URL, entries []BackendPingEntry) {
 			defer wg.Done()
 			ctx, cancel := context.WithTimeout(context.Background(), r.hub.backendTimeout)
 			defer cancel()
 
-			if err := r.hub.roomPing.SendPings(ctx, r, url, entries); err != nil {
+			if err := r.hub.roomPing.SendPings(ctx, r.id, url, entries); err != nil {
 				log.Printf("Error pinging room %s for active entries %+v: %s", r.id, entries, err)
 			}
 		}(urls[u], e)
 	}
-	return len(entries), &wg
+	return count, &wg
 }
 
 func (r *Room) publishRoomMessage(message *BackendRoomMessageRequest) {
-	if message == nil || message.Data == nil {
+	if message == nil || len(message.Data) == 0 {
 		return
 	}
 
@@ -1057,10 +1184,10 @@ func (r *Room) notifyInternalRoomDeleted() {
 		},
 	}
 
-	r.mu.Lock()
-	defer r.mu.Unlock()
+	r.mu.RLock()
+	defer r.mu.RUnlock()
 	for s := range r.internalSessions {
-		s.(*ClientSession).SendMessage(msg)
+		s.SendMessage(msg)
 	}
 }
 
