@@ -22,6 +22,7 @@
 package signaling
 
 import (
+	"context"
 	"fmt"
 )
 
@@ -34,4 +35,5 @@ type RoomSessions interface {
 	DeleteRoomSession(session Session)
 
 	GetSessionId(roomSessionId string) (string, error)
+	LookupSessionId(ctx context.Context, roomSessionId string, disconnectReason string) (string, error)
 }
