@@ -22,7 +22,9 @@
 package main
 
 import (
+	"context"
 	"crypto/tls"
+	"errors"
 	"flag"
 	"fmt"
 	"log"
@@ -34,6 +36,7 @@ import (
 	"runtime"
 	runtimepprof "runtime/pprof"
 	"strings"
+	"sync"
 	"syscall"
 	"time"
 
@@ -58,7 +61,7 @@ var (
 
 const (
 	defaultReadTimeout  = 15
-	defaultWriteTimeout = 15
+	defaultWriteTimeout = 30
 
 	initialMcuRetry = time.Second
 	maxMcuRetry     = time.Second * 16
@@ -91,6 +94,29 @@ func createTLSListener(addr string, certFile, keyFile string) (net.Listener, err
 	return tls.Listen("tcp", addr, &config)
 }
 
+type Listeners struct {
+	mu        sync.Mutex
+	listeners []net.Listener
+}
+
+func (l *Listeners) Add(listener net.Listener) {
+	l.mu.Lock()
+	defer l.mu.Unlock()
+
+	l.listeners = append(l.listeners, listener)
+}
+
+func (l *Listeners) Close() {
+	l.mu.Lock()
+	defer l.mu.Unlock()
+
+	for _, listener := range l.listeners {
+		if err := listener.Close(); err != nil {
+			log.Printf("Error closing listener %s: %s", listener.Addr(), err)
+		}
+	}
+}
+
 func main() {
 	log.SetFlags(log.Lshortfile)
 	flag.Parse()
@@ -103,6 +129,7 @@ func main() {
 	sigChan := make(chan os.Signal, 1)
 	signal.Notify(sigChan, os.Interrupt)
 	signal.Notify(sigChan, syscall.SIGHUP)
+	signal.Notify(sigChan, syscall.SIGUSR1)
 
 	if *cpuprofile != "" {
 		f, err := os.Create(*cpuprofile)
@@ -145,7 +172,7 @@ func main() {
 
 	signaling.RegisterStats()
 
-	natsUrl, _ := config.GetString("nats", "url")
+	natsUrl, _ := signaling.GetStringOptionWithEnv(config, "nats", "url")
 	if natsUrl == "" {
 		natsUrl = nats.DefaultURL
 	}
@@ -175,7 +202,7 @@ func main() {
 		}
 	}()
 
-	rpcServer, err := signaling.NewGrpcServer(config)
+	rpcServer, err := signaling.NewGrpcServer(config, version)
 	if err != nil {
 		log.Fatalf("Could not create RPC server: %s", err)
 	}
@@ -186,7 +213,7 @@ func main() {
 	}()
 	defer rpcServer.Close()
 
-	rpcClients, err := signaling.NewGrpcClients(config, etcdClient, dnsMonitor)
+	rpcClients, err := signaling.NewGrpcClients(config, etcdClient, dnsMonitor, version)
 	if err != nil {
 		log.Fatalf("Could not create RPC clients: %s", err)
 	}
@@ -198,7 +225,7 @@ func main() {
 		log.Fatal("Could not create hub: ", err)
 	}
 
-	mcuUrl, _ := config.GetString("mcu", "url")
+	mcuUrl, _ := signaling.GetStringOptionWithEnv(config, "mcu", "url")
 	mcuType, _ := config.GetString("mcu", "type")
 	if mcuType == "" && mcuUrl != "" {
 		log.Printf("WARNING: Old-style MCU configuration detected with url but no type, defaulting to type %s", signaling.McuTypeJanus)
@@ -214,9 +241,11 @@ func main() {
 		mcuRetryTimer := time.NewTimer(mcuRetry)
 	mcuTypeLoop:
 		for {
+			// Context should be cancelled on signals but need a way to differentiate later.
+			ctx := context.TODO()
 			switch mcuType {
 			case signaling.McuTypeJanus:
-				mcu, err = signaling.NewMcuJanus(mcuUrl, config)
+				mcu, err = signaling.NewMcuJanus(ctx, mcuUrl, config)
 				signaling.UnregisterProxyMcuStats()
 				signaling.RegisterJanusMcuStats()
 			case signaling.McuTypeProxy:
@@ -227,7 +256,7 @@ func main() {
 				log.Fatal("Unsupported MCU type: ", mcuType)
 			}
 			if err == nil {
-				err = mcu.Start()
+				err = mcu.Start(ctx)
 				if err != nil {
 					log.Printf("Could not create %s MCU: %s", mcuType, err)
 				}
@@ -248,7 +277,7 @@ func main() {
 					if config, err = goconf.ReadConfigFile(*configFlag); err != nil {
 						log.Printf("Could not read configuration from %s: %s", *configFlag, err)
 					} else {
-						mcuUrl, _ = config.GetString("mcu", "url")
+						mcuUrl, _ = signaling.GetStringOptionWithEnv(config, "mcu", "url")
 						mcuType, _ = config.GetString("mcu", "type")
 						if mcuType == "" && mcuUrl != "" {
 							log.Printf("WARNING: Old-style MCU configuration detected with url but no type, defaulting to type %s", signaling.McuTypeJanus)
@@ -300,7 +329,9 @@ func main() {
 		}
 	}
 
-	if saddr, _ := config.GetString("https", "listen"); saddr != "" {
+	var listeners Listeners
+
+	if saddr, _ := signaling.GetStringOptionWithEnv(config, "https", "listen"); saddr != "" {
 		cert, _ := config.GetString("https", "certificate")
 		key, _ := config.GetString("https", "key")
 		if cert == "" || key == "" {
@@ -328,14 +359,17 @@ func main() {
 					ReadTimeout:  time.Duration(readTimeout) * time.Second,
 					WriteTimeout: time.Duration(writeTimeout) * time.Second,
 				}
+				listeners.Add(listener)
 				if err := srv.Serve(listener); err != nil {
-					log.Fatal("Could not start server: ", err)
+					if !hub.IsShutdownScheduled() || !errors.Is(err, net.ErrClosed) {
+						log.Fatal("Could not start server: ", err)
+					}
 				}
 			}(address)
 		}
 	}
 
-	if addr, _ := config.GetString("http", "listen"); addr != "" {
+	if addr, _ := signaling.GetStringOptionWithEnv(config, "http", "listen"); addr != "" {
 		readTimeout, _ := config.GetInt("http", "readtimeout")
 		if readTimeout <= 0 {
 			readTimeout = defaultReadTimeout
@@ -359,26 +393,40 @@ func main() {
 					ReadTimeout:  time.Duration(readTimeout) * time.Second,
 					WriteTimeout: time.Duration(writeTimeout) * time.Second,
 				}
+				listeners.Add(listener)
 				if err := srv.Serve(listener); err != nil {
-					log.Fatal("Could not start server: ", err)
+					if !hub.IsShutdownScheduled() || !errors.Is(err, net.ErrClosed) {
+						log.Fatal("Could not start server: ", err)
+					}
 				}
 			}(address)
 		}
 	}
 
 loop:
-	for sig := range sigChan {
-		switch sig {
-		case os.Interrupt:
-			log.Println("Interrupted")
-			break loop
-		case syscall.SIGHUP:
-			log.Printf("Received SIGHUP, reloading %s", *configFlag)
-			if config, err := goconf.ReadConfigFile(*configFlag); err != nil {
-				log.Printf("Could not read configuration from %s: %s", *configFlag, err)
-			} else {
-				hub.Reload(config)
+	for {
+		select {
+		case sig := <-sigChan:
+			switch sig {
+			case os.Interrupt:
+				log.Println("Interrupted")
+				break loop
+			case syscall.SIGHUP:
+				log.Printf("Received SIGHUP, reloading %s", *configFlag)
+				if config, err := goconf.ReadConfigFile(*configFlag); err != nil {
+					log.Printf("Could not read configuration from %s: %s", *configFlag, err)
+				} else {
+					hub.Reload(config)
+					server.Reload(config)
+				}
+			case syscall.SIGUSR1:
+				log.Printf("Received SIGUSR1, scheduling server to shutdown")
+				hub.ScheduleShutdown()
+				listeners.Close()
 			}
+		case <-hub.ShutdownChannel():
+			log.Printf("All clients disconnected, shutting down")
+			break loop
 		}
 	}
 }
