@@ -22,9 +22,10 @@
 package signaling
 
 import (
+	"context"
 	"encoding/json"
 	"net/url"
-	"time"
+	"sync"
 )
 
 type Permission string
@@ -46,20 +47,16 @@ var (
 	}
 )
 
-type SessionIdData struct {
-	Sid       uint64
-	Created   time.Time
-	BackendId string
-}
-
 type Session interface {
+	Context() context.Context
 	PrivateId() string
 	PublicId() string
 	ClientType() string
 	Data() *SessionIdData
 
 	UserId() string
-	UserData() *json.RawMessage
+	UserData() json.RawMessage
+	ParsedUserData() (map[string]interface{}, error)
 
 	Backend() *Backend
 	BackendUrl() string
@@ -69,8 +66,25 @@ type Session interface {
 	GetRoom() *Room
 	LeaveRoom(notify bool) *Room
 
-	IsExpired(now time.Time) bool
 	Close()
 
 	HasPermission(permission Permission) bool
+
+	SendError(e *Error) bool
+	SendMessage(message *ServerMessage) bool
+}
+
+func parseUserData(data json.RawMessage) func() (map[string]interface{}, error) {
+	return sync.OnceValues(func() (map[string]interface{}, error) {
+		if len(data) == 0 {
+			return nil, nil
+		}
+
+		var m map[string]interface{}
+		if err := json.Unmarshal(data, &m); err != nil {
+			return nil, err
+		}
+
+		return m, nil
+	})
 }
