@@ -23,18 +23,16 @@ package signaling
 
 import (
 	"testing"
+
+	"github.com/stretchr/testify/assert"
 )
 
 func assertSessionHasPermission(t *testing.T, session Session, permission Permission) {
 	t.Helper()
-	if !session.HasPermission(permission) {
-		t.Errorf("Session %s doesn't have permission %s", session.PublicId(), permission)
-	}
+	assert.True(t, session.HasPermission(permission), "Session %s doesn't have permission %s", session.PublicId(), permission)
 }
 
 func assertSessionHasNotPermission(t *testing.T, session Session, permission Permission) {
 	t.Helper()
-	if session.HasPermission(permission) {
-		t.Errorf("Session %s has permission %s but shouldn't", session.PublicId(), permission)
-	}
+	assert.False(t, session.HasPermission(permission), "Session %s has permission %s but shouldn't", session.PublicId(), permission)
 }
