@@ -38,8 +38,10 @@ import (
 	"testing"
 	"time"
 
-	"github.com/golang-jwt/jwt/v4"
+	"github.com/golang-jwt/jwt/v5"
 	"github.com/gorilla/websocket"
+	"github.com/stretchr/testify/assert"
+	"github.com/stretchr/testify/require"
 )
 
 var (
@@ -64,7 +66,7 @@ func getWebsocketUrl(url string) string {
 }
 
 func getPubliceSessionIdData(h *Hub, publicId string) *SessionIdData {
-	decodedPublic := h.decodeSessionId(publicId, publicSessionName)
+	decodedPublic := h.decodePublicSessionId(publicId)
 	if decodedPublic == nil {
 		panic("invalid public session id")
 	}
@@ -98,7 +100,7 @@ func checkMessageType(message *ServerMessage, expectedType string) error {
 	case "message":
 		if message.Message == nil {
 			return fmt.Errorf("Expected \"%s\" message, got %+v", expectedType, message)
-		} else if message.Message.Data == nil || len(*message.Message.Data) == 0 {
+		} else if len(message.Message.Data) == 0 {
 			return fmt.Errorf("Received message without data")
 		}
 	case "room":
@@ -140,7 +142,7 @@ func checkReceiveClientMessageWithSenderAndRecipient(ctx context.Context, client
 	} else if err := checkMessageSender(client.hub, message.Message.Sender, senderType, hello); err != nil {
 		return err
 	} else {
-		if err := json.Unmarshal(*message.Message.Data, payload); err != nil {
+		if err := json.Unmarshal(message.Message.Data, payload); err != nil {
 			return err
 		}
 	}
@@ -170,7 +172,7 @@ func checkReceiveClientControlWithSenderAndRecipient(ctx context.Context, client
 	} else if err := checkMessageSender(client.hub, message.Control.Sender, senderType, hello); err != nil {
 		return err
 	} else {
-		if err := json.Unmarshal(*message.Control.Data, payload); err != nil {
+		if err := json.Unmarshal(message.Control.Data, payload); err != nil {
 			return err
 		}
 	}
@@ -225,9 +227,7 @@ type TestClient struct {
 func NewTestClientContext(ctx context.Context, t *testing.T, server *httptest.Server, hub *Hub) *TestClient {
 	// Reference "hub" to prevent compiler error.
 	conn, _, err := websocket.DefaultDialer.DialContext(ctx, getWebsocketUrl(server.URL), nil)
-	if err != nil {
-		t.Fatal(err)
-	}
+	require.NoError(t, err)
 
 	messageChan := make(chan []byte)
 	readErrorChan := make(chan error, 1)
@@ -238,8 +238,7 @@ func NewTestClientContext(ctx context.Context, t *testing.T, server *httptest.Se
 			if err != nil {
 				readErrorChan <- err
 				return
-			} else if messageType != websocket.TextMessage {
-				t.Errorf("Expect text message, got %d", messageType)
+			} else if !assert.Equal(t, websocket.TextMessage, messageType) {
 				return
 			}
 
@@ -266,13 +265,8 @@ func NewTestClient(t *testing.T, server *httptest.Server, hub *Hub) *TestClient
 
 	client := NewTestClientContext(ctx, t, server, hub)
 	msg, err := client.RunUntilMessage(ctx)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	if msg.Type != "welcome" {
-		t.Errorf("Expected welcome message, got %+v", msg)
-	}
+	require.NoError(t, err)
+	assert.Equal(t, "welcome", msg.Type)
 	return client
 }
 
@@ -311,12 +305,14 @@ func (c *TestClient) WaitForClientRemoved(ctx context.Context) error {
 	for {
 		found := false
 		for _, client := range c.hub.clients {
-			client.mu.Lock()
-			conn := client.conn
-			client.mu.Unlock()
-			if conn != nil && conn.RemoteAddr().String() == c.localAddr.String() {
-				found = true
-				break
+			if cc, ok := client.(*Client); ok {
+				cc.mu.Lock()
+				conn := cc.conn
+				cc.mu.Unlock()
+				if conn != nil && conn.RemoteAddr().String() == c.localAddr.String() {
+					found = true
+					break
+				}
 			}
 		}
 		if !found {
@@ -337,7 +333,7 @@ func (c *TestClient) WaitForClientRemoved(ctx context.Context) error {
 }
 
 func (c *TestClient) WaitForSessionRemoved(ctx context.Context, sessionId string) error {
-	data := c.hub.decodeSessionId(sessionId, publicSessionName)
+	data := c.hub.decodePublicSessionId(sessionId)
 	if data == nil {
 		return fmt.Errorf("Invalid session id passed")
 	}
@@ -378,9 +374,7 @@ func (c *TestClient) WriteJSON(data interface{}) error {
 }
 
 func (c *TestClient) EnsuerWriteJSON(data interface{}) {
-	if err := c.WriteJSON(data); err != nil {
-		c.t.Fatalf("Could not write JSON %+v: %s", data, err)
-	}
+	require.NoError(c.t, c.WriteJSON(data), "Could not write JSON %+v", data)
 }
 
 func (c *TestClient) SendHello(userid string) error {
@@ -395,18 +389,18 @@ func (c *TestClient) SendHelloV1(userid string) error {
 }
 
 func (c *TestClient) SendHelloV2(userid string) error {
-	now := time.Now()
-	return c.SendHelloV2WithTimes(userid, now, now.Add(time.Minute))
+	return c.SendHelloV2WithFeatures(userid, nil)
 }
 
-func (c *TestClient) SendHelloV2WithTimes(userid string, issuedAt time.Time, expiresAt time.Time) error {
-	userdata := map[string]string{
-		"displayname": "Displayname " + userid,
-	}
+func (c *TestClient) SendHelloV2WithFeatures(userid string, features []string) error {
+	now := time.Now()
+	return c.SendHelloV2WithTimesAndFeatures(userid, now, now.Add(time.Minute), features)
+}
 
+func (c *TestClient) CreateHelloV2TokenWithUserdata(userid string, issuedAt time.Time, expiresAt time.Time, userdata map[string]interface{}) (string, error) {
 	data, err := json.Marshal(userdata)
 	if err != nil {
-		c.t.Fatal(err)
+		return "", err
 	}
 
 	claims := &HelloV2TokenClaims{
@@ -414,7 +408,7 @@ func (c *TestClient) SendHelloV2WithTimes(userid string, issuedAt time.Time, exp
 			Issuer:  c.server.URL,
 			Subject: userid,
 		},
-		UserData: (*json.RawMessage)(&data),
+		UserData: data,
 	}
 	if !issuedAt.IsZero() {
 		claims.IssuedAt = jwt.NewNumericDate(issuedAt)
@@ -432,15 +426,29 @@ func (c *TestClient) SendHelloV2WithTimes(userid string, issuedAt time.Time, exp
 		token = jwt.NewWithClaims(jwt.SigningMethodRS256, claims)
 	}
 	private := getPrivateAuthToken(c.t)
-	tokenString, err := token.SignedString(private)
-	if err != nil {
-		c.t.Fatal(err)
+	return token.SignedString(private)
+}
+
+func (c *TestClient) CreateHelloV2Token(userid string, issuedAt time.Time, expiresAt time.Time) (string, error) {
+	userdata := map[string]interface{}{
+		"displayname": "Displayname " + userid,
 	}
 
+	return c.CreateHelloV2TokenWithUserdata(userid, issuedAt, expiresAt, userdata)
+}
+
+func (c *TestClient) SendHelloV2WithTimes(userid string, issuedAt time.Time, expiresAt time.Time) error {
+	return c.SendHelloV2WithTimesAndFeatures(userid, issuedAt, expiresAt, nil)
+}
+
+func (c *TestClient) SendHelloV2WithTimesAndFeatures(userid string, issuedAt time.Time, expiresAt time.Time, features []string) error {
+	tokenString, err := c.CreateHelloV2Token(userid, issuedAt, expiresAt)
+	require.NoError(c.t, err)
+
 	params := HelloV2AuthParams{
 		Token: tokenString,
 	}
-	return c.SendHelloParams(c.server.URL, HelloVersionV2, "", nil, params)
+	return c.SendHelloParams(c.server.URL, HelloVersionV2, "", features, params)
 }
 
 func (c *TestClient) SendHelloResume(resumeId string) error {
@@ -456,10 +464,14 @@ func (c *TestClient) SendHelloResume(resumeId string) error {
 }
 
 func (c *TestClient) SendHelloClient(userid string) error {
+	return c.SendHelloClientWithFeatures(userid, nil)
+}
+
+func (c *TestClient) SendHelloClientWithFeatures(userid string, features []string) error {
 	params := TestBackendClientAuthParams{
 		UserId: userid,
 	}
-	return c.SendHelloParams(c.server.URL, HelloVersionV1, "client", nil, params)
+	return c.SendHelloParams(c.server.URL, HelloVersionV1, "client", features, params)
 }
 
 func (c *TestClient) SendHelloInternal() error {
@@ -483,9 +495,7 @@ func (c *TestClient) SendHelloInternalWithFeatures(features []string) error {
 
 func (c *TestClient) SendHelloParams(url string, version string, clientType string, features []string, params interface{}) error {
 	data, err := json.Marshal(params)
-	if err != nil {
-		c.t.Fatal(err)
-	}
+	require.NoError(c.t, err)
 
 	hello := &ClientMessage{
 		Id:   "1234",
@@ -493,10 +503,10 @@ func (c *TestClient) SendHelloParams(url string, version string, clientType stri
 		Hello: &HelloClientMessage{
 			Version:  version,
 			Features: features,
-			Auth: HelloClientMessageAuth{
+			Auth: &HelloClientMessageAuth{
 				Type:   clientType,
 				Url:    url,
-				Params: (*json.RawMessage)(&data),
+				Params: data,
 			},
 		},
 	}
@@ -514,16 +524,14 @@ func (c *TestClient) SendBye() error {
 
 func (c *TestClient) SendMessage(recipient MessageClientMessageRecipient, data interface{}) error {
 	payload, err := json.Marshal(data)
-	if err != nil {
-		c.t.Fatal(err)
-	}
+	require.NoError(c.t, err)
 
 	message := &ClientMessage{
 		Id:   "abcd",
 		Type: "message",
 		Message: &MessageClientMessage{
 			Recipient: recipient,
-			Data:      (*json.RawMessage)(&payload),
+			Data:      payload,
 		},
 	}
 	return c.WriteJSON(message)
@@ -531,9 +539,7 @@ func (c *TestClient) SendMessage(recipient MessageClientMessageRecipient, data i
 
 func (c *TestClient) SendControl(recipient MessageClientMessageRecipient, data interface{}) error {
 	payload, err := json.Marshal(data)
-	if err != nil {
-		c.t.Fatal(err)
-	}
+	require.NoError(c.t, err)
 
 	message := &ClientMessage{
 		Id:   "abcd",
@@ -541,7 +547,7 @@ func (c *TestClient) SendControl(recipient MessageClientMessageRecipient, data i
 		Control: &ControlClientMessage{
 			MessageClientMessage: MessageClientMessage{
 				Recipient: recipient,
-				Data:      (*json.RawMessage)(&payload),
+				Data:      payload,
 			},
 		},
 	}
@@ -598,9 +604,7 @@ func (c *TestClient) SendInternalDialout(msg *DialoutInternalClientMessage) erro
 
 func (c *TestClient) SetTransientData(key string, value interface{}, ttl time.Duration) error {
 	payload, err := json.Marshal(value)
-	if err != nil {
-		c.t.Fatal(err)
-	}
+	require.NoError(c.t, err)
 
 	message := &ClientMessage{
 		Id:   "efgh",
@@ -608,7 +612,7 @@ func (c *TestClient) SetTransientData(key string, value interface{}, ttl time.Du
 		TransientData: &TransientDataClientMessage{
 			Type:  "set",
 			Key:   key,
-			Value: (*json.RawMessage)(&payload),
+			Value: payload,
 			TTL:   ttl,
 		},
 	}
@@ -722,6 +726,9 @@ func (c *TestClient) JoinRoomWithRoomSession(ctx context.Context, roomId string,
 	if err := checkMessageType(message, "room"); err != nil {
 		return nil, err
 	}
+	if message.Id != msg.Id {
+		return nil, fmt.Errorf("expected message id %s, got %s", msg.Id, message.Id)
+	}
 	return message, nil
 }
 
@@ -971,7 +978,7 @@ func (c *TestClient) RunUntilOffer(ctx context.Context, offer string) error {
 	}
 
 	var data map[string]interface{}
-	if err := json.Unmarshal(*message.Message.Data, &data); err != nil {
+	if err := json.Unmarshal(message.Message.Data, &data); err != nil {
 		return err
 	}
 
@@ -991,6 +998,10 @@ func (c *TestClient) RunUntilOffer(ctx context.Context, offer string) error {
 }
 
 func (c *TestClient) RunUntilAnswer(ctx context.Context, answer string) error {
+	return c.RunUntilAnswerFromSender(ctx, answer, nil)
+}
+
+func (c *TestClient) RunUntilAnswerFromSender(ctx context.Context, answer string, sender *MessageServerMessageSender) error {
 	message, err := c.RunUntilMessage(ctx)
 	if err != nil {
 		return err
@@ -1001,8 +1012,17 @@ func (c *TestClient) RunUntilAnswer(ctx context.Context, answer string) error {
 		return err
 	}
 
+	if sender != nil {
+		if err := checkMessageSender(c.hub, message.Message.Sender, sender.Type, &HelloServerMessage{
+			SessionId: sender.SessionId,
+			UserId:    sender.UserId,
+		}); err != nil {
+			return err
+		}
+	}
+
 	var data map[string]interface{}
-	if err := json.Unmarshal(*message.Message.Data, &data); err != nil {
+	if err := json.Unmarshal(message.Message.Data, &data); err != nil {
 		return err
 	}
 
@@ -1074,7 +1094,7 @@ func checkMessageInCallAll(message *ServerMessage, roomId string, inCall int) er
 		return fmt.Errorf("Expected participants update event for room %s, got %+v", roomId, message.Event.Update)
 	} else if !message.Event.Update.All {
 		return fmt.Errorf("Expected participants update event for all, got %+v", message.Event.Update)
-	} else if !bytes.Equal(*message.Event.Update.InCall, []byte(strconv.FormatInt(int64(inCall), 10))) {
+	} else if !bytes.Equal(message.Event.Update.InCall, []byte(strconv.FormatInt(int64(inCall), 10))) {
 		return fmt.Errorf("Expected incall flags %d, got %+v", inCall, message.Event.Update)
 	}
 	return nil
